import tempfile
from datetime import date
from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

from articat.config import ArticatConfig, ArticatMode
from articat.tests.utils import TestCatalog, TestFSArtifact

pytestmark = pytest.mark.datastore_emulated


def test_config_overwrite(uid: str):
    # Default Artifacts use default config
    assert ArticatConfig.fs_prod_prefix() in TestFSArtifact.config().fs_prod_prefix()

    custom_text = "SUPER_CUSTOM"

    # Get new FS paths:
    test_fs_prefix = Path(tempfile.mkdtemp(suffix=custom_text))
    for d in ("tmp", "dev", "prod"):
        test_fs_prefix.joinpath(d).mkdir(parents=True, exist_ok=True)

    custom_config = ArticatConfig(
        config_paths=(),
        config_dict=dict(
            ArticatConfig._config,
            fs={
                "tmp_prefix": test_fs_prefix.joinpath("tmp"),
                "dev_prefix": test_fs_prefix.joinpath("dev"),
                "prod_prefix": test_fs_prefix.joinpath("prod"),
            },
        ),
    )
    today = date.today()
    a1 = TestFSArtifact.partitioned(uid, today)
    a2 = TestFSArtifact.partitioned(f"{uid}_2", today, config=custom_config)

    assert custom_text not in a1.config().fs_prod_prefix()
    assert custom_text in a2.config().fs_prod_prefix()

    dummy_path = Path(tempfile.mktemp())
    dummy_path.touch()
    for i in (a1, a2):
        with i as a:
            a.stage(dummy_path)

    a1_dir = TestCatalog.get(a1.id, partition=today, model=TestFSArtifact).files_dir
    a2_dir = TestCatalog.get(a2.id, partition=today, model=TestFSArtifact).files_dir

    assert a1_dir and custom_text not in a1_dir
    assert a2_dir and custom_text in a2_dir


def test_config_env_variable(monkeypatch: MonkeyPatch):
    test_config = ArticatConfig._config
    test_config_mode = ArticatConfig.mode()
    assert test_config_mode == ArticatMode.gcp_datastore

    custom_config_path = Path(tempfile.mktemp())
    custom_config_path.write_text("[main]\n" "mode = local\n")

    try:
        with monkeypatch.context() as m:
            m.setenv("ARTICAT_CONFIG", custom_config_path.as_posix())
            ArticatConfig.register_config()
            assert ArticatConfig.mode() == ArticatMode.local
            assert ArticatConfig.mode() != test_config_mode
    finally:
        ArticatConfig._config = test_config



from __future__ import annotations

import logging
import os
import typing
from datetime import date, datetime
from os import environ
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

from google.cloud import datastore
from google.cloud.datastore.helpers import entity_to_protobuf
from pydantic import BaseModel, Extra, validator

from articat.config import ArticatConfig, ArticatMode, ConfigMixin
from articat.utils.datetime_utils import convert_to_datetime

if typing.TYPE_CHECKING:
    import pyspark

    from articat.catalog import Catalog

logger = logging.getLogger(__name__)


class Metadata(BaseModel):
    spark_schema: Optional[str] = None
    """String representation of spark schema"""
    schema_fields: Optional[List[str]] = None
    """List of fields in the schema"""
    arbitrary: Dict[str, Any] = {}
    """
    Any arbitrary metadata of your choice, can be embedded dicts, lists etc.
    Valid types: https://cloud.google.com/appengine/docs/standard/python/datastore/entities#Properties_and_value_types
    """
    description: Optional[str] = None
    """Description"""

    def add_spark_df_info(self, df: "pyspark.sql.DataFrame") -> "Metadata":
        """Add information from Spark Dataframe"""
        if self.arbitrary is not None:
            self.arbitrary.update({"nrow": df.count()})
        else:
            self.arbitrary = {"nrow": df.count()}
        self.schema_fields = df.columns
        self.spark_schema = df._jdf.schema().treeString()
        return self


class Arbitrary(BaseModel):
    """
    This is a documentation model, arbitrary dict isn't guaranteed to
    have any specific structure, but having these fields that we **try**
    to populate by default in one place is useful.
    """

    nrow: Optional[int]
    git_repo_url: Optional[str]
    git_head_hash: Optional[str]
    call_site_lineno: Optional[int]
    step_relfname: Optional[str]
    task_relfname: Optional[str]
    bq_stage_load_job_id: Optional[str]
    """BQ job id used for staging BQ data"""
    bq_job_id: Optional[str]
    """BQ job id used to produce the final BQ data"""
    execution_url: Optional[str]
    """URL to the execution that created this Artifact (could be for example GitHub Action)"""

    def get_update_dict(self) -> Dict[str, Union[int, str]]:
        """
        Returns dict with only the explicitly set fields.

        Useful to update arbitrary dict like this:

        a.metadata.arbitrary.update(Arbitrary(nrow=3).get_update_dict())

        ^ will only update the nrow field within arbitrary.
        """
        d = self.dict(exclude_unset=True)
        assert isinstance(d, dict)
        return d


# NOTE: see documentation below for the meaning of these fields
ID = str
Version = str
# NOTE: since `datetime` is also `date`, it's important that in
#       the union `datetime` precedes `date`.
Partition = Union[datetime, date]
EXECUTION_URL_ENV_NAME = "ARTICAT_EXECUTION_URL"
"""
If this env variable is set, the value will be recorded inside the arbitrary
dict. This could be for example URL to the GitHub Action workflow.
"""
T = TypeVar("T", bound="Artifact")


class Artifact(ConfigMixin, BaseModel):
    """Represents a single instance of a data Artifact"""

    class Config:
        """Core Artifact pydantic config"""

        extra = Extra.allow
        """
        Allow for extra properties to propagate into Artifact.

        This allows to parse any specialized Artifact as Artifact
        model, without "losing" extra properties.
        """
        underscore_attrs_are_private = True
        """Underscore attributes are private"""

    id: ID
    """Artifact ID, globally unique"""
    partition: Optional[Partition] = None
    """Artifact partition, akin to datetime "git commit hash" for the artifact"""
    metadata: Metadata = Metadata()
    """Partition's metadata"""
    version: Optional[Version] = None
    """Artifact version, akin to "git tag" for the artifact"""
    created: Optional[datetime] = None
    """Creation time"""
    _retire_entity: Optional[Mapping[str, Any]] = None
    # Note: this field is used to carry retired entity, it's not serialized
    _partition_str_format: ClassVar[str] = "%Y%m%dT%H%M%S"
    # string format for partition used in paths etc
    _config: Union[ArticatConfig, Type[ArticatConfig]] = ArticatConfig

    @validator("partition")
    def partition_must_be_datetime(cls, v: Optional[Partition]) -> Optional[datetime]:
        return None if v is None else convert_to_datetime(v)

    def open_browser(self) -> None:
        """Opens browser with the URL associated with this artifact. Best effort."""
        import webbrowser

        webbrowser.open(self.browser_url())

    def browser_url(self) -> str:
        """Returns URL of the resource associated with this artifact"""
        raise NotImplementedError()

    def build(self) -> "Artifact":
        """
        This method takes all the information provided in the current artifact
        and might decide to add more, for example for file based artifact,
        add file listing information, check if files exist, check size etc.
        """
        raise NotImplementedError(
            "You are trying to build abstract Artifact, "
            "use one of the concrete Artifact classes"
        )

    def _catalog(self) -> "Type[Catalog]":
        return self.config().catalog()

    def _exclude_private_fields(self) -> Set[str]:
        # TODO: remove since the config is in place
        return {f for f in self.__dict__ if f.startswith("_")}

    def _test_artifact_serialization(self) -> None:
        e = datastore.Entity()
        e.update(self.dict(exclude=self._exclude_private_fields()))
        entity_to_protobuf(e)

    def is_dev(self) -> bool:
        """Returns True if this artifact is in dev mode, False otherwise"""
        return self._is_dev_mode(self.id)

    @staticmethod
    def _is_dev_mode(id: Optional[ID]) -> bool:
        if not id:
            raise ValueError("id must be set")
        return id.startswith("_dev_")

    @staticmethod
    def _enforce_dev_mode(id: Optional[ID], dev: bool) -> Optional[str]:
        if id is None:
            return None
        if dev and not Artifact._is_dev_mode(id):
            return f"_dev_{id}"
        else:
            # NOTE: in the future we might log a warning here to use
            #       dev flag instead of id to specify the dev mode
            return id

    def __add_git_info(self) -> None:
        try:
            from articat.utils.utils import get_repo_and_hash

            repo_url, hash = get_repo_and_hash()
        except Exception as exc:
            logger.warning(f"Could not populate git info due to: {exc}")
            hash, repo_url = "UNKNOWN", "UNKNOWN"

        if hash.endswith("DIRTY") and not self.is_dev():
            # TODO (rav): make this an error eventually
            logger.warning(
                "Your git HEAD tree is dirty and you try to produce a production"
                " artifact, in the future this will raise an error!"
            )
        self.metadata.arbitrary.update(
            Arbitrary(git_repo_url=repo_url, git_head_hash=hash).get_update_dict()
        )

    def __add_exe_info(self) -> None:
        exe_url = os.environ.get(EXECUTION_URL_ENV_NAME)
        if exe_url is not None:
            self.metadata.arbitrary.update(
                Arbitrary(execution_url=exe_url.strip()).get_update_dict()
            )

    @classmethod
    def partitioned(
        cls: Type[T],
        id: ID,
        partition: Optional[Partition] = None,
        *,
        dev: bool = False,
        config: Optional[ArticatConfig] = None,
    ) -> T:
        """
        CTOR for a partitioned Artifact.

        :param id: Artifact ID.
        :param partition: Artifact partition.
        :param dev: mode mode flag.
        :param config: optional custom config for this Artifact.
        """
        a = cls(id=Artifact._enforce_dev_mode(id, dev), partition=partition)
        if config is not None:
            a._config = config
        return a

    @classmethod
    def versioned(
        cls: Type[T],
        id: ID,
        version: Version,
        *,
        dev: bool = False,
        config: Optional[ArticatConfig] = None,
    ) -> T:
        """
        CTOR for a versioned Artifact.

        :param id: Artifact ID.
        :param version: Artifact version.
        :param dev: mode mode flag.
        :param config: optional custom config for this Artifact.
        """
        a = cls(id=Artifact._enforce_dev_mode(id, dev), version=version)
        if config is not None:
            a._config = config
        return a

    def spec(self) -> Dict[str, Union[ID, Partition, Version]]:
        """
        Artifact spec is enough information to uniquely identify this
        artifact up to the partition/version. Useful to debug messages
        or recording dependencies.
        """
        r = self.dict(include={"id", "partition", "version"})
        assert isinstance(r, dict)
        return r

    def fetch(self: T, catalog: "Optional[Type[Catalog]]" = None) -> T:
        """
        When Artifact is used as a "spec", it doesn't have full information,
        use this method to fetch all the metadata from the Catalog.
        """
        assert isinstance(self, Artifact)

        catalog = catalog or self._catalog()
        return catalog.get(
            id=self.id,
            partition=self.partition,
            version=self.version,
            model=self.__class__,
        )

    def record_deps(self: T, *deps: "Artifact") -> T:
        """Record artifact's dependencies. Users can record deps using multiple
        calls to record_deps to appends deps"""
        assert isinstance(self, Artifact)
        d = self.metadata.arbitrary.get("deps", [])
        # dep spec will include just {id, partition, version},
        # having {id, partition} would actually be enough since
        # versioned artifacts still have unique partition, but
        # for completeness let's include version
        dep_spec = [i.spec() for i in deps]
        self.metadata.arbitrary.update(dict(deps=d + dep_spec))
        return self

    def __enter__(self: T) -> T:
        if (
            not hasattr(self, "__test__")
            and not self._is_dev_mode(self.id)
            and self.config().mode() != ArticatMode.local
        ):
            if "ARTICAT_PROD" not in environ:
                raise ValueError(
                    f"Looks like {self.spec()} would go to the production catalog,"
                    " but your environment is missing ARTICAT_PROD, aborting. If you truly"
                    " intend to write to production, set environment variable ARTICAT_PROD=1."
                )

        if self.created is not None:
            raise ValueError("You should not set created field by yourself")
        self.created = datetime.utcnow()
        if self.partition is None:
            self.partition = datetime.utcnow()
        if self.metadata is None:
            self.metadata = Metadata()
        if self.metadata.arbitrary is None:
            self.metadata.arbitrary = {}

        if self.is_dev():
            logger.warning(
                f"{self.spec()} is in the development mode, your outputs get deleted after 30 days"
            )

        assert self.metadata is not None
        assert self.created is not None
        assert self.partition is not None
        assert self.metadata.arbitrary is not None

        self.__add_git_info()
        self.__add_exe_info()

        # Check if the artifact metadata already exists:
        try:
            # NOTE: we use protected _lookup because we want to parse raw object later
            if self.version is not None:
                # NOTE: if version is set, we check by version
                e = next(
                    iter(self._catalog()._lookup(id=self.id, version=self.version))
                )
            else:
                e = next(
                    iter(
                        self._catalog()._lookup(
                            id=self.id,
                            partition_dt_start=self.partition,
                            partition_dt_end=self.partition,
                        )
                    )
                )
            if self.is_dev():
                logger.warning(
                    f"Dev mode will overwrite metadata for previous {self.id}"
                )
                object.__setattr__(self, "_retire_entity", e)
                return self
            raise ValueError("Catalog already has an entry for this artifact!")
        except StopIteration:
            return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if not exc_type:
            self._catalog().save(self.build())
            logger.info(f"Artifact {self.spec()} materialized at: {self.browser_url()}")
        object.__setattr__(self, "_retire_entity", None)


NoneArtifact = typing.cast(Artifact, object())
"""
This is a hack to allow keyword arguments in derived classes of Step
and fulfill LSP. You can use this object to provide default values for
input Artifacts, that will be injected by the orchestrator. We might
build it up a bit more later on, if the we see too many cryptic errors.
"""



from datetime import date
from pathlib import Path

from pytest import fixture

from articat.fs_artifact import FSArtifact
from articat.utils.notebook_utils import papermill_notebook

FIXTURE_DIR = Path(__file__).resolve().parent.joinpath("test_files")


@fixture
def dummy_notebook(tmpdir: str):
    return FIXTURE_DIR.joinpath("dummy_notebook.ipynb")


def test_papermill_notebook_default_params(caplog, dummy_notebook: Path):
    out = papermill_notebook(dummy_notebook)
    assert out.notebook_src == dummy_notebook
    assert out.notebook_exe.stat().st_size > 0
    assert out.notebook_html.stat().st_size > 0
    assert "foo artifact partition: None" in caplog.text


def test_papermill_notebook_custom_params(caplog, dummy_notebook: Path):
    out = papermill_notebook(
        dummy_notebook,
        params={
            "a": 3,
            "foo": FSArtifact.partitioned("foo", partition=date(1643, 1, 4), dev=True),
        },
    )
    assert out.notebook_src == dummy_notebook
    assert out.notebook_exe.stat().st_size > 0
    assert out.notebook_html.stat().st_size > 0
    assert "foo artifact partition: 1643-01-04" in caplog.text



from __future__ import annotations

import logging
from functools import lru_cache
from typing import Iterable, Optional, TypeVar

from google.cloud import datastore
from google.cloud.datastore import Client, Entity, Key

from articat.artifact import ID, Artifact, Metadata, Partition, Version
from articat.catalog import Catalog
from articat.utils.datetime_utils import convert_to_datetime

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Artifact)


class CatalogDatastore(Catalog):
    """GCP Datastore Articat Catalog"""

    @classmethod
    @lru_cache
    def _client(
        cls, project: Optional[str] = None, namespace: Optional[str] = None
    ) -> datastore.Client:
        project = project or cls.config().gcp_project()
        return datastore.Client(project=project, namespace=namespace)

    @classmethod
    def _dev_client(cls, project: Optional[str] = None) -> datastore.Client:
        project = project or cls.config().gcp_project()
        return cls._client(project=project, namespace="dev")

    @classmethod
    def _retired_client(cls, project: Optional[str] = None) -> datastore.Client:
        project = project or cls.config().gcp_project()
        return cls._client(project=project, namespace="retired")

    @classmethod
    def _lookup(
        cls,
        id: Optional[ID] = None,
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        dev: bool = False,
        client: Optional[Client] = None,
    ) -> Iterable[Entity]:
        id = Artifact._enforce_dev_mode(id, dev)
        if id and Artifact._is_dev_mode(id):
            msg_spec = dict(
                id=id,
                partition_dt_start=partition_dt_start,
                partition_dt_end=partition_dt_end,
                version=version,
            )
            logger.warning(
                f"Looking up dev artifact {msg_spec}, dev data gets deleted after 30 days"
            )
            client = client or cls._dev_client()
        elif id is None and dev:
            client = client or cls._dev_client()
        else:
            client = client or cls._client()
        logger.debug(
            f"Lookup of {id}, partition: [{partition_dt_start}, {partition_dt_end}), version: {version}, "
            f"metadata: {metadata}, limit: {limit}"
        )
        if metadata is not None:
            if metadata.spark_schema:
                raise ValueError(
                    "Spark schema lookup not supported, use schema_fields instead"
                )
            if metadata.description:
                raise ValueError("Description lookup not supported")
            if metadata.arbitrary:
                raise ValueError("Arbitrary lookup not supported")
        if metadata is not None and any(
            x is not None for x in [partition_dt_end, partition_dt_start]
        ):
            raise ValueError("Can't lookup based on both metadata and partition")
        if id:
            query = client.query(kind="Partition", ancestor=client.key("Artifact", id))
        else:
            query = client.query(kind="Partition")
        if partition_dt_start is not None or partition_dt_end is not None:
            if (
                partition_dt_start
                and partition_dt_end
                and partition_dt_start == partition_dt_end
            ):
                query.add_filter(
                    "partition", "=", convert_to_datetime(partition_dt_start)
                )
            else:
                if partition_dt_start is not None:
                    query.add_filter(
                        "partition", ">=", convert_to_datetime(partition_dt_start)
                    )
                if partition_dt_end is not None:
                    query.add_filter(
                        "partition", "<", convert_to_datetime(partition_dt_end)
                    )
                query.order = ["-partition"]
        if version:
            query.add_filter("version", "=", version)
        if metadata is not None:
            if metadata.schema_fields:
                for f in metadata.schema_fields:
                    query.add_filter("metadata.schema_fields", "=", f)
            if metadata.arbitrary is not None:
                for k, v in metadata.arbitrary.items():
                    query.add_filter(f"metadata.arbitrary.{k}", "=", v)
        yield from query.fetch(limit)

    @classmethod
    def _put_entity(cls, key: Key, artifact: Artifact, client: Client) -> None:
        catalog_entity = datastore.Entity(key)
        record = artifact.dict(exclude=artifact._exclude_private_fields())

        if artifact.metadata is not None:
            # This fields are not indexed, and therefor can't be match on
            # Note: due to https://github.com/googleapis/google-cloud-python/issues/1206
            metadata_entity = datastore.Entity(
                exclude_from_indexes=("spark_schema", "description", "arbitrary")
            )
            metadata_entity.update(artifact.metadata.dict())

            # Do not index arbitrary dict:
            arbitrary_entity = datastore.Entity(
                exclude_from_indexes=tuple(artifact.metadata.arbitrary.keys())
            )
            arbitrary_entity.update(artifact.metadata.arbitrary)
            metadata_entity["arbitrary"] = arbitrary_entity
            record["metadata"] = metadata_entity

        catalog_entity.update(record)
        client.put(catalog_entity)

    @classmethod
    def save(cls, artifact: Artifact) -> Artifact:
        """
        Saves a fully formed artifact into the Catalog. In most cases you
        should use artifact with a with-statement context, which builds
        and saves the artifact for you.
        :param artifact: artifact to save
        :return: saved Artifact
        """
        if not artifact.id:
            raise ValueError("id must be set to save an artifact")
        if artifact.partition is None:
            raise ValueError("partition must be set to save an artifact")
        if artifact.is_dev():
            client = cls._dev_client()
        else:
            client = cls._client()
        if artifact._retire_entity is not None:
            # We don't delete any metadata we copy it to the retired namespace
            re = artifact._retire_entity
            assert isinstance(re, Entity)
            r_art = artifact.__class__.parse_obj(re)
            retired_client = cls._retired_client()
            retired_key = retired_client.key(
                "Artifact", artifact.id, "Partition", re.id
            )
            cls._put_entity(retired_key, r_art, retired_client)
            key = client.key("Artifact", artifact.id, "Partition", re.id)
        else:
            key = client.key("Artifact", artifact.id, "Partition")
        # TODO (rav): make this a transaction at the key level
        cls._put_entity(key, artifact, client)
        return artifact



import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, Optional, Union

import fsspec

from articat.utils.typing import PathType


def to_pathlib(path: PathType) -> Path:
    """Coerce to pathlib Path"""
    if isinstance(path, str):
        return Path(path)
    else:
        return path


def fsspec_copyfile(
    src: str,
    dst: str,
    length: int = 0,
    src_compression: Optional[str] = None,
    dst_compression: Optional[str] = None,
) -> None:
    """
    Like `shutil.copyfile` but with support for `fsspec` paths
    and compression.
    """
    with fsspec.open(src, mode="rb", compression=src_compression) as fsrc, fsspec.open(
        dst, mode="wb", compression=dst_compression
    ) as fdst:
        shutil.copyfileobj(fsrc, fdst, length=length)


@contextmanager
def cwd(new_cwd: Union[str, Path]) -> Iterator[None]:
    """
    A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    Credit: https://gist.github.com/nottrobin/3d675653244f8814838a
    """

    prev_cwd = os.getcwd()
    try:
        os.chdir(new_cwd)
        yield
    finally:
        os.chdir(prev_cwd)



import logging
import sys
from datetime import date, datetime
from pathlib import Path
from typing import Literal, Optional

from articat.artifact import Artifact
from articat.bq_artifact import BQArtifact
from articat.config import ArticatConfig
from articat.fs_artifact import FSArtifact

logger = logging.getLogger(__name__)


class CLI:
    """Articat CLI interface"""

    @classmethod
    def open(
        cls,
        *,
        id: Optional[str] = None,
        partition: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        """
        Open URL associated with the Artifact.

        :param id: Artifact ID
        :param partition: Optional partition (YYYY-MM-DD)
        :param version: Optional version
        """
        artifact = Artifact(id=id, partition=partition, version=version).fetch()
        if hasattr(artifact, "files_pattern"):
            artifact = FSArtifact.parse_obj(artifact)
        elif hasattr(artifact, "table_id"):
            artifact = BQArtifact.parse_obj(artifact)
        else:
            raise ValueError(f"Don't know how to open {artifact}")
        logger.info(f"Opening artifact: {artifact.spec()} via {artifact.browser_url()}")
        artifact.open_browser()

    @staticmethod
    def lookup(
        *,
        id: Optional[str] = None,
        partition: Optional[str] = None,
        version: Optional[str] = None,
        limit: Optional[int] = None,
        no_trunc: bool = False,
        dev: bool = False,
        format: Literal["csv", "json"] = "json",
    ) -> None:
        """
        Does RS Catalog lookup and prints results. All parameters are exact, case sensitive match.
        :param id: artifact id
        :param partition: artifact partition date (YYYY-MM-DD)
        :param version: artifact version
        :param limit: limit the number of results
        :param no_trunc: whether to print full metadata or a short version (default: True)
        :param dev: switch to dev mode (default: False)
        :param format: output format, default to json, can also be csv
        """
        assert format in ("json", "csv"), "format has to be one of: json, csv"
        partition_dt = date.fromisoformat(partition) if partition else None
        csv_result = []
        json_exclude = (
            None
            if no_trunc
            else {
                "metadata": ...,
            }
        )
        for e in ArticatConfig.catalog().lookup(
            id=id,
            partition_dt_start=partition_dt
            or datetime.fromisoformat(date.min.isoformat()),
            partition_dt_end=partition_dt,
            version=str(version) if version else None,
            limit=limit,
            model=Artifact,
            dev=dev,
        ):
            if format == "json":
                print(e.json(exclude=json_exclude), flush=True)  # type: ignore[arg-type]
            else:
                assert format == "csv"
                csv_result.append(e.dict(exclude=dict(metadata=...)))
        if len(csv_result) > 0:
            import pandas as pd

            pd.json_normalize(csv_result, sep="_").to_csv(sys.stdout, index=False)

    @staticmethod
    def dump(
        *,
        local_path: str,
        id: str,
        version: str,
        description: str,
        dev: bool = False,
        _test_artifact: Optional[FSArtifact] = None,
    ) -> FSArtifact:
        """
        Ad-hoc dump of local data into the Catalog, creates a versioned FSArtifact.

        :param local_path: local path to dump, can be a file or a directory (directories are recursively uploaded)
        :param id: id of the artifact
        :param version: version
        :param description: meaningful description
        :param dev: whether to use dev mode
        :return: Created FSArtifact
        """
        artifact = _test_artifact or FSArtifact.versioned(
            id=id, version=version, dev=dev
        )
        with artifact as artifact:
            artifact.stage(Path(local_path))
            artifact.metadata.description = description
        return artifact



from datetime import datetime

import pytest

from articat.bq_artifact import BQArtifact


def test_versioned_not_supported():
    with pytest.raises(
        NotImplementedError, match="Versioned BigQuery Artifact is not supported"
    ):
        BQArtifact.versioned("foo", version="0.1.0")


def test_bq_client_is_lazy():
    assert BQArtifact.bq_client


def test_partitioned_must_be_date():
    assert BQArtifact.partitioned("foo")
    with pytest.raises(
        ValueError, match="Partition resolution for BQ artifact must be a day"
    ):
        BQArtifact.partitioned("foo", partition=datetime.utcnow())



from __future__ import annotations

import logging
from datetime import date
from typing import Any, Iterable, Mapping, Optional, Type, TypeVar, Union, overload

from articat.artifact import ID, Artifact, Metadata, Partition, Version
from articat.config import ArticatConfig, ConfigMixin

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Artifact)


class Catalog(ConfigMixin):
    """RS Data Catalog"""

    _config: Union[Type[ArticatConfig], ArticatConfig] = ArticatConfig

    @classmethod
    @overload
    def get(
        cls,
        id: ID,
        *,
        version: Optional[Version],
        dev: bool = False,
    ) -> Artifact:
        ...

    @classmethod
    @overload
    def get(
        cls,
        id: ID,
        *,
        partition: Optional[Partition],
        dev: bool = False,
    ) -> Artifact:
        ...

    @classmethod
    @overload
    def get(cls, id: ID, *, dev: bool = False) -> Artifact:
        ...

    @classmethod
    @overload
    def get(cls, id: ID, *, model: Type[T], dev: bool = False) -> T:
        ...

    @classmethod
    @overload
    def get(
        cls,
        id: ID,
        *,
        model: Type[T],
        version: Optional[Version] = None,
        partition: Optional[Partition] = None,
        dev: bool = False,
    ) -> T:
        ...

    # NOTE: no return type due to: https://github.com/python/mypy/issues/3737
    @classmethod
    def get(  # type: ignore[no-untyped-def]
        cls,
        id: ID,
        *,
        version: Optional[Version] = None,
        partition: Optional[Partition] = None,
        model: Type[Artifact] = Artifact,
        dev: bool = False,
    ):
        """
        Get a data artifact given its id. You should specify partition or
        version for deterministic retrieval. Model class can be specified
        to parse the result as a specific model.

        If you do not specify partition, you will get the latest partition,
        which might not be the latest created artifact with that ID. If
        you want to fetch the latest artifact with a given ID, sort by
        created field, example:

        ```
        sorted(Catalog.lookup("foo"), key=lambda x: x.created)[-1]
        ```

        :param id: artifact id
        :param version: optional artifact version
        :param partition: optional datetime/date partition
        :param model: optional model class (e.g. FSArtifact)
        :type dev: whether to use dev mode (default False)
        """
        if not partition and not version:
            a = cls.latest_partition(id, model=model, dev=dev)
            assert isinstance(a, Artifact)
            logger.warning(
                "No version or partition specified, fetching latest partition. "
                "This should be avoided in non development pipelines. "
                f"Fetched {a.id}, partition: {a.partition} version: {a.version}"
            )
            return a
        try:
            return next(
                iter(
                    cls.lookup(
                        id=id,
                        partition_dt_start=partition,
                        partition_dt_end=partition,
                        version=version,
                        limit=1,
                        model=model,
                        dev=dev,
                    )
                )
            )
        except StopIteration as e:
            req = dict(id=id, partition=partition, version=version, dev=dev)
            raise ValueError(f"Can't find requested artifact {req}") from e

    @classmethod
    @overload
    def latest_partition(cls, id: ID, *, dev: bool = False) -> Artifact:
        ...

    @classmethod
    @overload
    def latest_partition(cls, id: ID, *, model: Type[T], dev: bool = False) -> T:
        ...

    # NOTE: no return type due to: https://github.com/python/mypy/issues/3737
    @classmethod
    def latest_partition(  # type: ignore[no-untyped-def]
        cls, id: ID, *, model: Type[Artifact] = Artifact, dev: bool = False
    ):
        try:
            return next(
                iter(
                    cls.lookup(
                        id=id,
                        partition_dt_start=date.min,
                        limit=1,
                        model=model,
                        dev=dev,
                    )
                )
            )
        except StopIteration as e:
            raise ValueError(f"Can't find requested artifact {id}") from e

    @classmethod
    @overload
    def lookup(
        cls,
        id: Optional[ID] = None,
        *,
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        dev: bool = False,
    ) -> Iterable[Artifact]:
        ...

    @classmethod
    @overload
    def lookup(
        cls,
        id: Optional[ID] = None,
        *,
        model: Type[T],
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        dev: bool = False,
    ) -> Iterable[T]:
        ...

    # NOTE: no return type due to: https://github.com/python/mypy/issues/3737
    @classmethod
    def lookup(  # type: ignore[no-untyped-def]
        cls,
        id: Optional[ID] = None,
        *,
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        model: Type[Artifact] = Artifact,
        dev: bool = False,
    ):
        """
        Lookup/search atop Catalog. You can search based on id, partitions, version
        and most metadata. Keep in mind that substring matches, case-insensitive matches,
        or so-called full-text search are not supported, if that is your
        use case match as much as you can and do further matching locally.
        You can't lookup by id + partition/version and metadata at the same time.

        Examples:

        ```
        Catalog.lookup() # returns an iterable over whole catalog
        # returns an iterable over all partitions/versions within my-id:
        Catalog.lookup("my-id")
        # return an iterable over partitions since yesterday, ordered desc by partition
        # aka latest is first in the iterable:
        Catalog.lookup("my-id", partition_dt_start=yesterday)
        # lookup my-id with version 0.1.2
        Catalog.lookup("my-id", version="0.1.2")
        ```

        :param id: Id of the artifact/dataset group
        :param partition_dt_start: start datetime/date partition (including)
        :param partition_dt_end: end datetime/date partition (excluding)
        :param version: version to match
        :param metadata: metadata to match
        :param limit: limit number of results
        :param model: what model to use to parse the results
        :param dev: whether to use dev mode
        :return: Iterable[<model>]
        """
        assert issubclass(model, Artifact)
        yield from (
            model.parse_obj(r)
            for r in cls._lookup(
                id=id,
                partition_dt_start=partition_dt_start,
                partition_dt_end=partition_dt_end,
                version=version,
                metadata=metadata,
                limit=limit,
                dev=dev,
            )
        )

    @classmethod
    def _lookup(
        cls,
        id: Optional[ID] = None,
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        dev: bool = False,
    ) -> Iterable[Mapping[str, Any]]:
        raise NotImplementedError()

    @classmethod
    def save(cls, artifact: Artifact) -> Artifact:
        """
        Saves a fully formed artifact into the Catalog. In most cases you
        should use artifact with a with-statement context, which builds
        and saves the artifact for you.
        :param artifact: artifact to save
        :return: saved Artifact
        """
        raise NotImplementedError()

    @classmethod
    def to_dataframe(
        cls,
        dev: bool = False,
        limit: Optional[int] = None,
        include_arbitrary: bool = True,
    ) -> Any:
        """
        Return DataFrame representation of the Catalog.

        Use `dev=True` to get the development view of the Catalog. Use `limit` to reduce the
        size of the output. `include_arbitrary=False` to avoid expanding arbitrary field.
        """
        try:
            import pandas as pd
        except ImportError:
            logger.exception("to_dataframe requires pandas, install pandas ...")
            raise

        exclude = None if include_arbitrary else dict(metadata=dict(arbitrary=...))
        catalog_dicts = list(
            i.dict(exclude=exclude)  # type: ignore[arg-type]
            for i in cls.lookup(dev=dev, limit=limit, model=Artifact)
        )
        return pd.json_normalize(catalog_dicts, sep="_")



import os
from contextlib import contextmanager
from datetime import date, timedelta
from functools import lru_cache
from random import shuffle
from typing import TYPE_CHECKING, ClassVar, Iterator, Optional, Type, TypeVar

import fsspec
from google.cloud import datastore

from articat.artifact import ID, Partition, Version
from articat.catalog import Catalog
from articat.catalog_datastore import CatalogDatastore
from articat.config import ArticatConfig
from articat.fs_artifact import FSArtifact

T = TypeVar("T", bound="TestFSArtifactMixin")


class TestCatalog(CatalogDatastore):
    """Datastore Catalog used for tests"""

    @classmethod
    @lru_cache
    def _client(
        cls, project: str = "foobar", namespace: Optional[str] = None
    ) -> datastore.Client:
        os.environ["DATASTORE_EMULATOR_HOST"] = "127.0.0.1:8099"
        return datastore.Client(project=project, namespace=namespace)


if TYPE_CHECKING:
    # NOTE: we want to use TestFSArtifactMixin across multiple
    #       test classes, and flavours of FSArtifact
    BASE_CLASS = FSArtifact
else:
    BASE_CLASS = object


class TestFSArtifactMixin(BASE_CLASS):
    # Note: we overwrite str format in tests to avoid partition path conflicts
    #       for versioned outputs
    _partition_str_format: ClassVar[str] = "%Y%m%dT%H%M%S%f"

    @classmethod
    def _catalog(cls) -> "Type[Catalog]":
        return TestCatalog

    @classmethod
    @contextmanager
    def dummy_versioned_ctx(
        cls: Type[T], uid: ID, version: Version, dev: bool = False
    ) -> Iterator[T]:
        with cls.versioned(uid, version, dev=dev) as a:
            with fsspec.open(a.joinpath("output.txt"), "w") as f:
                f.write("ala ma kota")
            a.files_pattern = f"{a.staging_file_prefix}/output.txt"
            yield a

    @classmethod
    def write_dummy_versioned(
        cls: Type[T], uid: ID, version: Version, dev: bool = False
    ) -> T:
        with cls.dummy_versioned_ctx(uid, version, dev=dev) as a:
            ...
        return a

    @classmethod
    @contextmanager
    def dummy_partitioned_ctx(
        cls: Type[T], uid: ID, partition: Partition, dev: bool = False
    ) -> Iterator[T]:
        with cls.partitioned(uid, partition=partition, dev=dev) as a:
            with fsspec.open(a.joinpath("output.txt"), "w") as f:
                f.write("ala ma kota")
            assert ArticatConfig.fs_tmp_prefix() in a.staging_file_prefix
            a.files_pattern = f"{a.staging_file_prefix}/output.txt"
            yield a

    @classmethod
    def write_dummy_partitioned(
        cls: Type[T], uid: ID, partition: Partition, dev: bool = False
    ) -> T:
        with cls.dummy_partitioned_ctx(uid, partition, dev=dev) as a:
            ...
        return a


class TestFSArtifact(TestFSArtifactMixin, FSArtifact):
    """FSArtifact used for tests"""

    __test__: ClassVar[bool] = False
    """
    This field prevents pytest from interpreting TestFSArtifact as
    a test class. Artifact can also check for the existence of
    __test__ to check if it's running the the test context.
    """


def write_a_couple_of_partitions(uid: ID, n: int) -> None:
    # we shuffle the partitions to further tests ordering etc
    ns = list(range(n))
    shuffle(ns)
    for i in ns:
        dt = date.today() - timedelta(days=i)
        TestFSArtifact.write_dummy_partitioned(uid, dt)
        # write some unrelated dataset
        TestFSArtifact.write_dummy_partitioned(f"{uid}1234", dt)



from typing import Any, Dict, Optional

import fsspec
from gcsfs import GCSFileSystem

from articat.fs_artifact import FSArtifact
from articat.utils.notebook_utils import papermill_notebook
from articat.utils.path_utils import fsspec_copyfile, to_pathlib
from articat.utils.typing import PathType


class NotebookArtifact(FSArtifact):
    """
    A simple Notebook flavour of the FSArtifact. Each notebook artifact,
    should hold a **single** notebook output, and the notebook should be
    staged via the `stage_notebook` method.
    """

    def browser_url(self, filename: str = "output.html") -> str:
        """
        Returns the URL of the HTML output of the artifact. If you have
        a custom HTML report specify `filename`.
        """
        if self.main_dir.startswith("gs://"):
            auth_url_root = "https://storage.cloud.google.com"
            path = self.main_dir.lstrip("gs://")
            return f"{auth_url_root}/{path}/{filename}"
        else:
            return super().browser_url()

    def stage_rendered_notebook(
        self, report_path: PathType, filename: Optional[str] = "output.html"
    ) -> "NotebookArtifact":
        """
        Stages pre-rendered notebook. Default filename is `output.html`, if you want to
        keep original filename, set `filename` to None.
        """
        report_path_real = to_pathlib(report_path)
        html_output = self.joinpath(filename or report_path_real.name)
        fsspec_copyfile(report_path_real.as_posix(), html_output)

        fs = fsspec.open(html_output).fs
        if isinstance(fs, GCSFileSystem):
            # NOTE: we want to be able to open the HTML output in the browser
            #       from GCS, and thus need to adjust the content-type
            fs.setxattrs(path=html_output, content_type="text/html")
        return self

    def stage_notebook(
        self, notebook_path: PathType, params: Dict[str, Any] = {}
    ) -> "NotebookArtifact":
        """
        Allows to easily stage a notebook within this artifact. It saves the original
        notebook (with original name), the executed notebook with _executed suffix, and
        the HTML output in `output.html`.
        """
        nb_out = papermill_notebook(to_pathlib(notebook_path), params)
        fsspec_copyfile(
            nb_out.notebook_src.as_posix(), self.joinpath(nb_out.notebook_src.name)
        )
        fsspec_copyfile(
            nb_out.notebook_exe.as_posix(), self.joinpath(nb_out.notebook_exe.name)
        )
        return self.stage_rendered_notebook(nb_out.notebook_html)



import logging
import os
import re
import time
import uuid
from functools import lru_cache
from multiprocessing.pool import ThreadPool
from pathlib import Path
from types import TracebackType
from typing import ClassVar, List, Optional, Type, TypeVar

import fsspec
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from gcsfs import GCSFileSystem

from articat.artifact import Artifact

logger = logging.getLogger(__name__)

T = TypeVar("T", bound="FSArtifact")


class FSArtifact(Artifact):
    """
    File/object based Artifact. Metadata associated with files/objects
    stored on file system/object store. Provides utilities for staging
    data. Location of temp/dev/prod results can be configured.

    Example:

        with FSArtifact.partitioned("id", date.today(), dev=True) as fsa:
            with fsspec.open(fsa.joinpath("out.txt"), "w") as fd:
                fd.write("hello world")
    """

    # Regex of the files/objects produced by this Artifact, needs to be
    # set by the user
    files_pattern: Optional[str] = None
    # Represents the "root" of the artifacts file structure, populated
    # automatically
    files_dir: Optional[str] = None

    # PRIVATE fields:
    _file_prefix: Optional[str] = None
    _fs_scheme_regex: ClassVar[str] = r"^.*?://"

    @property
    def staging_file_prefix(self) -> str:
        """
        This is your staging location for outputs. When you are done make sure to
        update `files_pattern` with a pattern/glob that includes all the outputs,
        within the `staging_file_prefix`.
        """
        if not self._file_prefix:
            raise ValueError(
                "staging_file_prefix is unset, if you are creating a new "
                "artifact make sure to use it within with-statement "
                "context. Artifacts retrieved from Catalog do not "
                "have staging_file_prefix set, and you are probably looking "
                "for `files_pattern` field."
            )
        else:
            return self._file_prefix

    @property
    def main_dir(self) -> str:
        """Returns a "main path" to the artifact.

        The main path is where all the files are stored, user has
        and full freedom on the structure inside the main directory,
        and `files_pattern` is a pattern within the main path.
        """
        # to not worry about presence of scheme, we remove the prefix
        # and add it back
        if self.files_dir:
            return self.files_dir
        # TODO: remove this code path
        # This is the legacy handling of the main_dir, where we gather
        # main dir from the files_pattern
        assert self.files_pattern is not None
        if self.is_dev():
            no_prefix = self.files_pattern.replace(self.config().fs_dev_prefix(), "")
            return self.config().fs_dev_prefix() + "/".join(no_prefix.split("/")[:4])
        else:
            no_prefix = self.files_pattern.replace(self.config().fs_prod_prefix(), "")
            return self.config().fs_prod_prefix() + "/".join(no_prefix.split("/")[:3])

    def joinpath(self, *parts: str) -> str:
        """Helper to easily construct paths within the Artifact"""
        # NOTE: ATTOW, fsspec.open still requires str as path, and since
        #       fsspec is our de facto FS interface, here we still return
        #       string as path (instead of for example pathlib.Path)
        #       Related: https://github.com/intake/filesystem_spec/issues/114
        if self._file_prefix is not None:
            return os.path.join(self.staging_file_prefix, *parts)
        else:
            return os.path.join(self.main_dir, *parts)

    @staticmethod
    @lru_cache
    def _worker_pool() -> ThreadPool:
        # Note: this worker pool is used for file ops
        return ThreadPool()

    @staticmethod
    def _publish_file(
        fs: AbstractFileSystem, src: str, staging_prefix: str, dst_prefix: str
    ) -> str:
        staging_prefix = re.sub(FSArtifact._fs_scheme_regex, "", staging_prefix)
        dst_prefix = re.sub(FSArtifact._fs_scheme_regex, "", dst_prefix)
        if staging_prefix not in src:
            raise ValueError(
                f"Looks like staged file pattern is outside of the staging space: {staging_prefix} "
                f"you can only publish data from within staging prefix."
            )
        dst = src.replace(staging_prefix, dst_prefix)
        # We copy not move, to keep the staging intact
        logger.debug(f"Will publish `{src}` to `{dst}`")
        try:
            fs.cp(src, dst)
        except FileNotFoundError as e:
            if isinstance(fs, GCSFileSystem):
                if fs.isdir(src):
                    logger.debug(
                        f"Skipping file not found for a fake directory on GCS: {src}"
                    )
                    return dst
            raise e
        return dst

    def build(self) -> "FSArtifact":
        """
        Builds this artifact. In most cases you should use artifact within with-statement
        context which takes care of building and saving artifacts for you, and that's the
        recommended usage.
        """
        # This is mostly to test that metadata doesn't have any illegal types before
        # we copy any files etc
        self._test_artifact_serialization()
        if not self.files_pattern:
            std_output = f"{self.staging_file_prefix}/**"
            logger.warning(f"Will publish whole staging directory: {std_output}")
            self.files_pattern = std_output
        fs = fsspec.filesystem(self._get_protocol(self.files_pattern))

        # Note: this could potentially be problematic for very large number of
        # files, as in thousands of files, when that becomes and issue, we
        # can optimize it.
        start = time.time()
        output_files: List[str] = fs.glob(self.files_pattern)
        logger.debug(f"Temp output glob took: {time.time() - start} s")

        if len(output_files) == 0:
            raise ValueError(
                f"There are no files in the output `{self.files_pattern}`, aborting"
            )

        # Next:
        # * we copy files from staging area
        # * add a _SUCCESS marker file
        # * dump all the metadata about this artifact in a _MANIFEST.json file
        actual_prefix = self._get_file_prefix(tmp=False)
        # NOTE: if local FS, we need to create directory structures
        self._create_if_local(actual_prefix)
        start = time.time()
        self._worker_pool().map(
            lambda p: self._publish_file(
                fs, p, self.staging_file_prefix, actual_prefix
            ),
            output_files,
        )
        self.files_pattern = self.files_pattern.replace(
            self.staging_file_prefix, actual_prefix
        )
        self.files_dir = actual_prefix
        logger.debug(f"Output files published in: {time.time() - start}s")
        if fs.exists(f"{actual_prefix}/_MANIFEST.json"):
            raise ValueError(
                "Manifest file `_MANIFEST.json` already present in the staged files - this is not supported"
            )
        with fs.open(f"{actual_prefix}/_MANIFEST.json", "w") as m:
            m.write(self.json(exclude=self._exclude_private_fields()))
        try:
            # truncate false, means if the file already exists do not
            # truncate it, which is what we want. If the file
            # exists, `touch` on AbstractFS and GCSFS throws
            # NotImplementedError, in which case we just double check
            # if the SUCCESS marker exist, and raise if it not
            fs.touch(f"{actual_prefix}/_SUCCESS", truncate=False)
        except NotImplementedError:
            assert fs.exists(
                f"{actual_prefix}/_SUCCESS"
            ), "We could not create a SUCCESS marker and it does not exist"
        return self

    def _get_file_prefix(self, tmp: bool = True) -> str:
        if not self.created:
            raise ValueError("created must be set")
        if not self.id:
            raise ValueError("ID must be set")

        partition_path = self.created.strftime(Artifact._partition_str_format)
        if tmp:
            return f"{self.config().fs_tmp_prefix()}/{self.id}/{partition_path}/{uuid.uuid4()}"
        elif self.is_dev():
            return f"{self.config().fs_dev_prefix()}/{self.id}/{partition_path}/{uuid.uuid4()}"
        else:
            return f"{self.config().fs_prod_prefix()}/{self.id}/{partition_path}/{uuid.uuid4()}"

    @staticmethod
    def _get_protocol(path: Optional[str]) -> str:
        if not path:
            raise ValueError("Path should not be None")
        else:
            return path.split(":")[0] if ":" in path else "file"

    @classmethod
    def _create_if_local(cls, path: str) -> None:
        fs = fsspec.filesystem(cls._get_protocol(path))
        if isinstance(fs, LocalFileSystem):
            fs.mkdirs(path, exist_ok=True)

    def __enter__(self: T) -> T:
        r = super().__enter__()
        assert isinstance(r, FSArtifact)

        if not r._file_prefix:
            # This kind of setter flavour is required by pydantic for private fields
            object.__setattr__(r, "_file_prefix", r._get_file_prefix(tmp=True))

        # NOTE: if local FS, we need to create directory structures
        r._create_if_local(r.staging_file_prefix)
        return r

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        object.__setattr__(self, "_file_prefix", None)

    def browser_url(self) -> str:
        if self.main_dir.startswith("gs://"):
            path = self.main_dir.lstrip("gs://")
            return f"https://console.cloud.google.com/storage/browser/{path}"
        else:
            return self.main_dir

    def stage(self, local_path: Path) -> "FSArtifact":
        """
        Dumps local file/directory into a FSArtifact. `local_path` is a file or
        a directory.
        """
        if not local_path.exists():
            raise ValueError(f"`{local_path.as_posix()} does not exists`")
        fs = fsspec.get_filesystem_class(
            FSArtifact._get_protocol(self.staging_file_prefix)
        )()
        if local_path.is_dir():
            fs.upload(
                local_path.as_posix(),
                self.staging_file_prefix,
                recursive=True,
            )
        else:
            fs.upload(local_path.as_posix(), self.joinpath(local_path.name))
        return self



from datetime import date, datetime

import fsspec
import pytest
from _pytest.monkeypatch import MonkeyPatch
from dateutil.tz import UTC

from articat.artifact import ID, Artifact, Metadata
from articat.fs_artifact import FSArtifact
from articat.tests.utils import (
    TestCatalog,
    TestFSArtifact,
    write_a_couple_of_partitions,
)


def test_large_arbitrary_entry(uid: ID) -> None:
    more_than_1500 = "f" * 2000
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.metadata.arbitrary.update(dict(foo=more_than_1500))

    a = TestCatalog.get(uid, version="0.1.0", model=TestFSArtifact)
    assert a.metadata.arbitrary.get("foo") == more_than_1500


def test_artifact_parse_extra_prop(uid: ID) -> None:
    TestFSArtifact.write_dummy_partitioned(uid, date.today())
    a_back = TestCatalog.get(uid, partition=date.today(), model=Artifact)
    assert getattr(a_back, "files_pattern")


def test_write_to_prod_on_non_prod_env_fails(uid: ID, monkeypatch: MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.delattr(TestFSArtifact, "__test__")
        with pytest.raises(
            ValueError, match="your environment is missing ARTICAT_PROD"
        ):
            TestFSArtifact.write_dummy_versioned(uid, "0.1.0")


def test_catalog_deps__happy_path(uid: ID) -> None:
    dep1 = TestFSArtifact.partitioned(
        "dep1", partition=datetime.today().replace(tzinfo=UTC)
    )
    dep2 = TestFSArtifact.versioned("dep2", version="0.1.1")
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.record_deps(dep1, dep2)

    a = TestCatalog.latest_partition(uid, model=TestFSArtifact)
    deps = a.metadata.arbitrary.get("deps")
    assert isinstance(deps, list)
    assert len(deps) == 2
    assert TestFSArtifact.spec(dep1) in deps
    assert TestFSArtifact.spec(dep2) in deps


def test_catalog_deps__dup_dep(uid: ID) -> None:
    dep1 = TestFSArtifact.partitioned(
        "dep1", partition=datetime.today().replace(tzinfo=UTC)
    )
    dep2 = TestFSArtifact.versioned("dep2", version="0.1.1")
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.record_deps(dep1)
        a.record_deps(dep2)
        # NOTE: if the user specifies the same dep multiple times
        #       it will end up in the deps multiple times
        a.record_deps(dep1)

    a = TestCatalog.latest_partition(uid, model=TestFSArtifact)
    deps = a.metadata.arbitrary.get("deps")
    assert isinstance(deps, list)
    assert len(deps) == 3
    assert TestFSArtifact.spec(dep1) in deps
    assert TestFSArtifact.spec(dep2) in deps


def test_catalog_fails_on_preexisting_manifest_file(uid: ID) -> None:
    with pytest.raises(
        ValueError, match="Manifest file `_MANIFEST.json` already present"
    ):
        with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
            with fsspec.open(f"{a.staging_file_prefix}/_MANIFEST.json", "w") as f:
                f.write("foobar")
            a.files_pattern = None


def test_dev_mode(uid: ID) -> None:
    write_a_couple_of_partitions(f"_dev_{uid}", 1)
    r = TestCatalog.get(uid, partition=date.today(), dev=True, model=FSArtifact)
    assert TestFSArtifact.config().fs_dev_prefix() in (r.files_pattern or "")
    assert "final_output" not in (r.files_pattern or "final_output")


def test_dev_mode_in_out(uid: ID) -> None:
    TestFSArtifact.write_dummy_versioned(uid, "0.1.0", dev=True)
    with pytest.raises(ValueError, match="Can't find requested artifact"):
        TestCatalog.get(uid, version="0.1.0")
    assert TestCatalog.get(uid, version="0.1.0", dev=True) is not None


def test_dev_mode_dont_dup_prefix(uid: ID) -> None:
    TestFSArtifact.write_dummy_versioned(f"_dev_{uid}", "0.1.0", dev=True)
    assert TestCatalog.get(f"_dev_{uid}", version="0.1.0") is not None
    assert TestCatalog.get(uid, version="0.1.0", dev=True) is not None
    assert TestCatalog.get(f"_dev_{uid}", version="0.1.0", dev=True) is not None


def test_fails_before_file_copy_on_invalid_types_in_metadata(uid: ID) -> None:
    with pytest.raises(ValueError, match="Unknown protobuf attr"):
        with TestFSArtifact.partitioned(uid) as a:
            a.metadata = Metadata(arbitrary={"a": date.today()})


def test_metadata_supports_embedded_dicts_lists(uid: ID) -> None:
    today = date.today()
    nested_metadata = {"a": {"foo": "bar", "baz": [1, 2, 3]}}
    with TestFSArtifact.dummy_partitioned_ctx(uid, today) as a:
        a.metadata = Metadata(arbitrary={"a": {"foo": "bar", "baz": [1, 2, 3]}})

    roundback_metadata = TestCatalog.get(uid, partition=today).metadata
    assert roundback_metadata and roundback_metadata.arbitrary == nested_metadata



"""
data artifact catalog
"""

from articat.artifact import Artifact
from articat.bq_artifact import BQArtifact
from articat.catalog_local import CatalogLocal
from articat.config import ArticatConfig
from articat.fs_artifact import FSArtifact
from articat.notebook_artifact import NotebookArtifact

__all__ = [
    "ArticatConfig",
    "Artifact",
    "BQArtifact",
    "CatalogLocal",
    "FSArtifact",
    "NotebookArtifact",
]



import logging
import uuid
from datetime import date, timedelta
from functools import lru_cache
from types import TracebackType
from typing import Any, ClassVar, Optional, Type

from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from articat.artifact import ID, Arbitrary, Artifact, Partition, Version
from articat.config import ArticatConfig

logger = logging.getLogger(__name__)


class BQArtifact(Artifact):
    """
    BigQuery artifact for the RS Catalog. It creates natively partitioned table in BigQuery.
    Partitioned table doc: https://cloud.google.com/bigquery/docs/partitioned-tables#ingestion_time.
    The partition filter is required on the queries.

    This artifact is currently in "beta", we want to figure out how do we want to
    use BigQuery and what works best.
    """

    _bq_partition_date_format: ClassVar[str] = "%Y%m%d"
    """BQ uses this format to specify the day partitioning table"""
    _staging_table: Optional[str] = None
    """Staging table the user should write data to"""
    table_id: str = ""
    """
    BQ table id, this exists when you retrieve the artifact and right
    after the write. Format: <PROJECT_ID>.<DATASET_ID>.<TABLE_ID>.
    """

    @staticmethod
    @lru_cache
    def bq_client(project: Optional[str] = None, **kwargs: Any) -> bigquery.Client:
        """MVP BQ client for BQ Artifacts"""
        # TODO (rav): add support for custom GCP project/etc
        return bigquery.Client(project=project, **kwargs)

    @classmethod
    def versioned(
        cls,
        id: ID,
        version: Version,
        *,
        dev: bool = False,
        config: Optional[ArticatConfig] = None,
    ) -> "BQArtifact":
        """Versioned BQ artifacts are not supported"""
        raise NotImplementedError("Versioned BigQuery Artifact is not supported")

    @classmethod
    def partitioned(
        cls,
        id: ID,
        partition: Optional[Partition] = date.today(),
        *,
        dev: bool = False,
        config: Optional[ArticatConfig] = None,
    ) -> "BQArtifact":
        """Partitioned BQ artifact, the partition must be a date"""
        assert partition is not None
        # TODO (rav): add support for up to an hour resolution
        if partition.resolution != timedelta(days=1):
            raise ValueError("Partition resolution for BQ artifact must be a day")
        return super().partitioned(id=id, partition=partition, dev=dev, config=config)

    def __enter__(self) -> "BQArtifact":
        r: BQArtifact = super().__enter__()
        assert r.partition

        if not r.created:
            raise ValueError("created must be set")
        if not r.id:
            raise ValueError("ID must be set")

        partition_str = r.created.strftime(Artifact._partition_str_format)
        # NOTE: this style of assignment is required: https://github.com/samuelcolvin/pydantic/issues/655
        object.__setattr__(
            r,
            "_staging_table",
            f"{r.config().gcp_project()}.{r.config().bq_tmp_dataset}.{r.id}_{partition_str}_{uuid.uuid4()}",
        )

        partition_str = r.partition.strftime(BQArtifact._bq_partition_date_format)
        if r.is_dev():
            r.table_id = f"{r.config().gcp_project()}.{r.config().bq_dev_dataset()}.{r.id}${partition_str}"
        else:
            r.table_id = f"{r.config().gcp_project()}.{r.config().bq_prod_dataset()}.{r.id}${partition_str}"
        logger.debug(f"Final data of {r.spec()} will end up in {r.table_id}")
        return r

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        # NOTE: this style of set is required: https://github.com/samuelcolvin/pydantic/issues/655
        object.__setattr__(self, "_staging_table", None)

    def build(self) -> "BQArtifact":
        assert self._staging_table, "Staging table must be set"
        assert self.table_id, "Destination table id must be set"
        assert self.partition, "Partition must be set"

        bq_client = self.bq_client(self.config().gcp_project())

        staged_table = bq_client.get_table(self._staging_table)

        try:
            tbl = bq_client.get_table(self.table_id.split("$")[0])
            if (
                tbl.description != self.metadata.description
                or tbl.schema != staged_table.schema
            ):
                # TODO: update only what is actually needed
                tbl.description = self.metadata.description
                tbl.schema = staged_table.schema
                bq_client.update_table(table=tbl, fields=["schema", "description"])
            logger.debug(f"Found table {tbl}")
        except NotFound:
            logger.info(
                f"Table {self.table_id} does not exist, will create a new partitioned table"
            )
            tbl = bigquery.Table(
                self.table_id.split("$")[0],
                schema=bq_client.get_table(self._staging_table).schema,
            )
            tbl.description = self.metadata.description
            tbl.time_partitioning = bigquery.TimePartitioning(
                type_=bigquery.TimePartitioningType.DAY,
                # TODO (rav): what should be the default partition expiration, if not
                #             specified it will inherit the dataset expiration
                # expiration_ms=...,
            )
            tbl.require_partition_filter = True
            bq_client.create_table(tbl, exists_ok=False)

        if self.is_dev():
            write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE
        else:
            write_disposition = bigquery.job.WriteDisposition.WRITE_EMPTY

        job_config = bigquery.CopyJobConfig(
            create_disposition=bigquery.job.CreateDisposition.CREATE_IF_NEEDED,
            write_disposition=write_disposition,
        )

        cp_job = bq_client.copy_table(
            sources=self._staging_table,
            destination=self.table_id,
            job_config=job_config,
        )
        cp_job_result = cp_job.result()

        destination_table = bq_client.get_table(self.table_id)
        logger.info(f"Loaded {destination_table.num_rows} rows to {self.table_id}")
        self.metadata.arbitrary.update(
            Arbitrary(bq_job_id=cp_job_result.job_id).get_update_dict()
        )
        return self

    def browser_url(self) -> str:
        project_id, dataset_id, table_id = self.table_id.split(".")
        bq_url_root = "https://console.cloud.google.com/bigquery"
        return f"{bq_url_root}?project={project_id}&p={project_id}&d={dataset_id}&page=table&t={table_id}"



from datetime import date, datetime
from typing import Union


def convert_to_datetime(dt: Union[datetime, date]) -> datetime:
    # NOTE: isinstance(date.today(), datetime) is True
    return (
        datetime.combine(dt, datetime.min.time())
        if not isinstance(dt, datetime)
        else dt
    )



from __future__ import annotations

from functools import partial, update_wrapper
from typing import Callable, Generic, Optional, Type, TypeVar

T_OUT = TypeVar("T_OUT")
T_DECO = TypeVar("T_DECO")


class class_or_instance_method(Generic[T_OUT]):
    """
    Instance/class method decorator. If available uses the instance,
    if called on class uses the class.

    Note: you can't easily implement a reliable "class property",
    without running into edge cases that will evaluate the property,
    for example help(klass).

    Example:

        class Foo:
            _bar: int = 42

            def __init__(self):
                self._bar = 128

            @class_or_instance_method
            def bar(self_or_cls):
                return self_or_cls._bar

        Foo.bar() #42
        Foo().bar() #128
    """

    def __init__(self, method: Callable[[T_DECO], T_OUT]):
        # NOTE: method appears unbound, so we don't need to
        #       worry about weakref here.
        self.org_method = method

    def __get__(
        self, instance: Optional[T_DECO], cls: Type[T_DECO]
    ) -> Callable[[], T_OUT]:
        self_or_cls = instance or cls
        wrapper = partial(self.org_method, self_or_cls)
        return update_wrapper(wrapper, self.org_method)



import logging
import os
import tempfile
import warnings
from configparser import ConfigParser
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, Type, Union

from articat.utils.class_or_instance_method import class_or_instance_method

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from articat.catalog import Catalog


class ArticatMode(str, Enum):
    local = "local"
    gcp_datastore = "gcp_datastore"


class ArticatConfig:
    """
    Articat config handler. Allows to read configuration from config files,
    and config dictionaries. See python's `configparser` for context. This
    class supports "global"/default config and instances of configuration
    which you can supply to Artifact to have custom configured Artifact
    objects, to for example materialise certain Artifacts in custom locations.
    """

    # NOTE: the order in this list matters, and defines the override order
    default_config_paths = [
        Path.home().joinpath(".config", "articat", "articat.cfg").as_posix(),
        Path.cwd().joinpath("articat.cfg").as_posix(),
    ]
    _config: ConfigParser = ConfigParser()

    def __init__(
        self,
        config_paths: Sequence[str] = [],
        config_dict: Mapping[str, Mapping[str, Any]] = {},
    ) -> None:
        self._config = self._read_config(
            config_paths=config_paths, config_dict=config_dict
        )

    @staticmethod
    def _read_config(
        config_paths: Sequence[str], config_dict: Mapping[str, Mapping[str, Any]]
    ) -> ConfigParser:
        config = ConfigParser()
        read = config.read(config_paths)
        if len(read) == 0:
            logger.warning(f"No configuration files found, searched in {config_paths}")
        config.read_dict(config_dict)
        return config

    @classmethod
    def register_config(
        cls,
        config_paths: Optional[Sequence[str]] = None,
        config_dict: Mapping[str, Mapping[str, Any]] = {},
    ) -> "Type[ArticatConfig]":
        """
        Register configuration from config paths and config dictionary. Config
        paths are read in order, `config_dict` is applied after config paths.
        This methods registers "global"/default configuration, use constructor
        to get an instance of a config. You can also specify the configuration
        location via env variable: ARTICAT_CONFIG.

        `config_paths` resolution order:
         * function argument
         * `ARTICAT_CONFIG` env variable
         * default paths
        """
        if config_paths is None:
            env_config_path = os.environ.get("ARTICAT_CONFIG")
            if env_config_path is not None:
                config_paths = [env_config_path]
            else:
                config_paths = cls.default_config_paths
        cls._config = cls._read_config(
            config_paths=config_paths, config_dict=config_dict
        )
        return cls

    @class_or_instance_method
    def mode(self) -> ArticatMode:
        """
        Articat mode. Currently supported: `local`, `gcp`.
        Defaults to: `local`
        """
        return ArticatMode(self._config.get("main", "mode", fallback=ArticatMode.local))

    @class_or_instance_method
    def catalog(self) -> "Type[Catalog]":
        """Returns the Catalog implementation for given mode"""
        if self.mode() == ArticatMode.local:
            from articat.catalog_local import CatalogLocal

            return CatalogLocal
        elif self.mode() == ArticatMode.gcp_datastore:
            from articat.catalog_datastore import CatalogDatastore

            return CatalogDatastore
        else:
            raise ValueError(f"Unknown catalog for mode: {self.mode}")

    @class_or_instance_method
    def local_db_dir(self) -> str:
        """
        Location of the local DB, `local` mode only.
        Defaults to: ~/.config/articat/local
        """
        return self._config.get(
            "main",
            "local_db_dir",
            fallback=Path.home().joinpath(".config", "articat", "local").as_posix(),
        )

    @class_or_instance_method
    def gcp_project(self) -> str:
        """Google Cloud Platform (GCP) project, for `gcp` mode only"""
        return self._config["gcp"]["project"]

    @class_or_instance_method
    def bq_prod_dataset(self) -> str:
        """BigQuery (BQ) production dataset, for `gcp` mode only"""
        return self._config["bq"]["prod_dataset"]

    @class_or_instance_method
    def bq_tmp_dataset(self) -> str:
        """BigQuery (BQ) temp dataset, for `gcp` mode only"""
        return self._config["bq"]["tmp_dataset"]

    @class_or_instance_method
    def bq_dev_dataset(self) -> str:
        """BigQuery (BQ) development dataset, for `gcp` mode only"""
        return self._config["bq"]["dev_dataset"]

    @classmethod
    def _set_option(
        cls, c: ConfigParser, section: str, option: str, value: Any
    ) -> None:
        if not c.has_section(section):
            c.add_section(section)
        c.set(section, option, value)

    @class_or_instance_method
    def fs_tmp_prefix(self) -> str:
        """File system (FS) temporary/staging location"""
        try:
            return self._config["fs"]["tmp_prefix"]
        except KeyError:
            r = tempfile.mkdtemp(suffix="artifact_temp_dir_")
            self._set_option(self._config, "fs", "tmp_prefix", r)
            warnings.warn(
                f"FSArtifact temp directory not configured, assuming local mode, using temp directory: {r}"
            )
            return r

    @class_or_instance_method
    def fs_dev_prefix(self) -> str:
        """File system (FS) development location"""
        try:
            return self._config["fs"]["dev_prefix"]
        except KeyError:
            r = Path.cwd().joinpath(".articat_catalog", "dev").as_posix()
            self._set_option(self._config, "fs", "dev_prefix", r)
            warnings.warn(
                f"FSArtifact development directory not configured, assuming local mode, using cwd: {r}"
            )
            return r

    @class_or_instance_method
    def fs_prod_prefix(self) -> str:
        """File system (FS) production/final location"""
        try:
            return self._config["fs"]["prod_prefix"]
        except KeyError:
            r = Path.cwd().joinpath(".articat_catalog", "prod").as_posix()
            self._set_option(self._config, "fs", "prod_prefix", r)
            warnings.warn(
                f"FSArtifact production directory not configured, assuming local mode, using cwd: {r}"
            )
            return r


class ConfigMixin:
    """ArticatConfig mixin/trait"""

    _config: Union[Type[ArticatConfig], ArticatConfig]

    @class_or_instance_method
    def config(self) -> Union[Type[ArticatConfig], ArticatConfig]:
        """Get Articat config object"""
        return self._config



from __future__ import annotations

import dbm
from datetime import datetime
from hashlib import md5
from pathlib import Path
from typing import Any, Iterable, Mapping, Optional

from articat.artifact import ID, Artifact, Metadata, Partition, Version
from articat.catalog import Catalog
from articat.utils.datetime_utils import convert_to_datetime


class CatalogLocal(Catalog):
    """
    This is local flavour of the Catalog. This implementation is purely for
    exploration, test, presentation etc. This implementation is inefficient
    by design. Do not use for production use cases. This implementation is
    not thread-safe.
    """

    @classmethod
    def _get_db(cls) -> dbm._Database:
        # NOTE: this is inefficient, but that's fine for the purpose of the local mode
        Path(cls.config().local_db_dir()).mkdir(parents=True, exist_ok=True)
        return dbm.open(
            Path(cls.config().local_db_dir()).joinpath("store").as_posix(), flag="c"
        )

    @classmethod
    def _compute_key(cls, artifact: Artifact) -> str:
        h = md5()
        h.update(artifact.id.encode())
        if artifact.version:
            h.update(artifact.version.encode())
        if artifact.partition:
            assert isinstance(artifact.partition, datetime)
            h.update(artifact.partition.isoformat().encode())
        return h.hexdigest()

    @classmethod
    def _lookup(
        cls,
        id: Optional[ID] = None,
        partition_dt_start: Optional[Partition] = None,
        partition_dt_end: Optional[Partition] = None,
        version: Optional[Version] = None,
        metadata: Optional[Metadata] = None,
        limit: Optional[int] = None,
        dev: bool = False,
    ) -> Iterable[Mapping[str, Any]]:
        if metadata is not None:
            raise ValueError("Local mode doesn't currently support lookup by metadata")
        result = []
        with cls._get_db() as db:
            for k in db.keys():
                a = Artifact.parse_raw(db[k])
                if id is not None:
                    if a.id != id:
                        continue
                if version is not None:
                    if a.version != version:
                        continue
                if (
                    partition_dt_start is not None or partition_dt_end is not None
                ) and a.partition is not None:
                    if partition_dt_start == partition_dt_end and partition_dt_start:
                        if a.partition != convert_to_datetime(partition_dt_start):
                            continue
                    else:
                        if partition_dt_start is not None:
                            if a.partition < convert_to_datetime(partition_dt_start):
                                continue
                        if partition_dt_end is not None:
                            if a.partition >= convert_to_datetime(partition_dt_end):
                                continue
                result.append(a.dict())
        yield from sorted(
            result, key=lambda x: (x["partition"], x["version"]), reverse=True
        )

    @classmethod
    def save(cls, artifact: Artifact) -> Artifact:
        with cls._get_db() as db:
            if artifact._retire_entity is not None:
                del db[cls._compute_key(Artifact.parse_obj(artifact._retire_entity))]
            db[cls._compute_key(artifact)] = artifact.json()
        return artifact



import logging
import tempfile
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Mapping, cast

from google.cloud.bigquery._helpers import _datetime_to_json
from nbconvert import HTMLExporter
from nbformat.v4 import nbjson
from papermill import execute_notebook
from papermill.translators import PythonTranslator, papermill_translators

from articat.artifact import Artifact

logger = logging.getLogger(__name__)


@dataclass
class PapermillOutput:
    """Value class to hold output of the papermill_notebook function"""

    notebook_src: Path
    """Holds the path to the source notebook"""
    notebook_exe: Path
    """Holds the path to the papermill executed notebook"""
    notebook_html: Path
    """Holds the path to the HTML export of the executed notebook"""


def papermill_notebook(
    notebook_path: Path, params: Dict[str, Any] = {}
) -> PapermillOutput:
    """
    Uses papermill to execute notebook with potential parameters. Read the papermill documentation
    for a tutorial on the parameter handling. It saves executed notebook and an HTML report in a
    temporary directory. The original notebook is untouched.
    """
    tmp_dir = Path(tempfile.mkdtemp())
    logger.debug(
        f"Working/temp directory for papermill on {notebook_path} with {params} will be {tmp_dir}"
    )

    assert notebook_path.name.endswith(
        ".ipynb"
    ), f"Invalid notebook file suffix in {notebook_path.name}"
    notebook_exe_path = tmp_dir.joinpath(
        notebook_path.name.replace(".ipynb", "_executed.ipynb")
    )

    logger.debug(f"Temp executed notebook will be saved in {notebook_exe_path}")

    nb_node = execute_notebook(
        input_path=notebook_path,
        output_path=notebook_exe_path,
        parameters=params,
        log_output=True,
    )

    html_exporter = HTMLExporter()
    html_body, resources = html_exporter.from_notebook_node(nb_node)

    notebook_html_path = tmp_dir.joinpath("output.html")
    notebook_html_path.write_text(html_body)

    out = PapermillOutput(
        notebook_src=notebook_path,
        notebook_exe=notebook_exe_path,
        notebook_html=notebook_html_path,
    )

    logger.info(
        f"Notebook {out.notebook_src} execution saved in {out.notebook_exe}, HTML export in {out.notebook_html}"
    )
    return out


class RSNBJsonEncoder(nbjson.BytesEncoder):  # type: ignore[misc]
    """
    RS flavour of the Notebook JsonEncoder.

    This is required for notebook API to be able to serialize
    Artifact objects, this is being called by papermill.
    """

    def default(self, obj: Any) -> str:
        if isinstance(obj, Artifact):
            # NOTE: we really only need the spec, it's assumed that
            #       the notebook will call fetch() on the artifacts
            #       to retrieve metadata at the runtime.
            return cast(str, super().encode(obj.spec()))
        elif isinstance(obj, datetime):
            return cast(str, _datetime_to_json(obj))
        elif isinstance(obj, Mapping):
            return cast(str, super().encode(dict(obj)))
        else:
            return cast(str, super().default(obj))


nbjson.BytesEncoder = RSNBJsonEncoder
# NOTE: here we overwrite the default BytesEncoder in the notebook internals,
#       afaiu there's no mechanism to do it gracefully, thus we need this
#       uncivilized, brute method above


class RSPythonTranslator(PythonTranslator):  # type: ignore[misc]
    """
    RS flavour of the papermill translator.

    My current understanding of the Translator: it takes a python object,
    and should return a stringified code that would recreate it the
    notebook cell. We need a custom Translator to teach it how to recreate
    the Artifact object.
    """

    @classmethod
    def translate_artifact(cls, val: Artifact) -> str:
        # NOTE: here we need to translate an existing Artifact object into
        #       a line of code (str) that would recreate it in the notebook.
        #       It is assumed that the notebook should have all the necessary
        #       imports, otherwise it will fail at the runtime with undefined
        #       symbol. It's a fair assumption, given that the user should
        #       provide default Artifact spec in the parameter cell.
        spec = cls.translate(val.spec())
        return f"{val.__class__.__name__}(**{spec})"

    @classmethod
    def translate(cls, val: Any) -> str:
        if isinstance(val, Artifact):
            return cls.translate_artifact(val)
        else:
            return cast(str, super().translate(val))


papermill_translators.register("python", RSPythonTranslator)
# NOTE: here we overwrite the default python translator



import re
import tempfile
from pathlib import Path

import fsspec
import pytest

from articat.artifact import ID
from articat.fs_artifact import FSArtifact
from articat.tests.utils import TestCatalog, TestFSArtifact


def test_fs_stage(uid: ID, tmpdir: str) -> None:
    tmpdir_path = Path(tmpdir).joinpath("foo")
    tmpdir_path.write_text("foo")

    with TestFSArtifact.versioned(uid, version="foo") as a:
        a.stage(tmpdir_path)
    assert Path(a.joinpath("foo")).read_text() == "foo"


def test_fs_artifact_join_paths_works(uid: ID) -> None:
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        assert a._file_prefix is not None
        p = a.joinpath("data.dat")
        with fsspec.open(p, "w") as f:
            f.write("ala ma kota")
        a.files_pattern = None
        assert p == f"{a.staging_file_prefix}/data.dat"
    assert a._file_prefix is None
    p = a.joinpath("data.dat")
    assert p == f"{a.main_dir}/data.dat"
    assert Path(p).read_text() == "ala ma kota"


def test_catalog_main_path_is_valid(uid: ID) -> None:
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.files_pattern = None
    a = TestCatalog.latest_partition(uid, model=TestFSArtifact)
    assert "**" not in a.main_dir
    assert a.files_pattern
    assert a.files_pattern.endswith("/**")
    assert a.files_pattern.startswith(a.main_dir)
    assert a.main_dir == a.files_pattern.replace("/**", "")


def test_catalog_main_path_is_valid_in_dev_mode(uid: ID) -> None:
    TestFSArtifact.write_dummy_versioned(f"_dev_{uid}", "0.1.0")
    a = TestCatalog.latest_partition(f"_dev_{uid}", model=TestFSArtifact)
    assert "**" not in a.main_dir
    assert a.files_pattern
    assert a.files_pattern.endswith("/output.txt")
    assert a.files_pattern.startswith(a.main_dir)
    assert a.main_dir == a.files_pattern.replace("/output.txt", "")


def test_catalog_no_files_pattern_includes_everything(uid: ID) -> None:
    with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.files_pattern = None
    output = TestCatalog.get(uid, version="0.1.0", model=FSArtifact).files_pattern
    assert output is not None
    assert output.endswith("/**")


def test_fail_on__missing_output(uid: ID) -> None:
    with pytest.raises(ValueError, match="There are no files in the output"):
        with TestFSArtifact.partitioned(uid) as art:
            out_dir = Path(tempfile.mkdtemp())
            art.files_pattern = f"{out_dir.as_posix()}/*"


def test_publish_file_works_for_local() -> None:
    fs = fsspec.filesystem("file")
    prefix = f"{tempfile.mkdtemp()}"
    fs.touch(f"{prefix}/sth.txt")
    dst_prefix = f"{tempfile.mkdtemp()}"
    TestFSArtifact._publish_file(fs, f"{prefix}/sth.txt", prefix, dst_prefix)
    assert Path(f"{dst_prefix}/sth.txt").exists()


def test_fs_scheme_regex() -> None:
    assert (
        re.sub(TestFSArtifact._fs_scheme_regex, "", "gs://test-sth/foobar")
        == "test-sth/foobar"
    )
    assert (
        re.sub(TestFSArtifact._fs_scheme_regex, "", "gs://test-sth/foobar://fds[.*]")
        == "test-sth/foobar://fds[.*]"
    )



from datetime import date, datetime, time, timedelta
from pathlib import Path
from typing import List, cast

import pytest
from _pytest.monkeypatch import MonkeyPatch
from dateutil.tz import UTC

from articat.artifact import EXECUTION_URL_ENV_NAME, ID, Metadata
from articat.fs_artifact import FSArtifact
from articat.tests.utils import (
    TestCatalog,
    TestFSArtifact,
    write_a_couple_of_partitions,
)

# Tests in this module require Datastore emulator, and by default won't run
# unless pytest has been told that the emulator is in fact available
# via `--datastore_emulated`. Out Github CI has Datastore emulator enabled.
# To run these tests locally:
# docker run --rm -p 8099:8099 ravwojdyla/datastore_emulator:0.1.0
# pytest --datastore_emulated
pytestmark = pytest.mark.datastore_emulated


def test_simple_happy_path_catalog(uid: ID) -> None:
    yesterday = datetime.utcnow() - timedelta(days=1)
    today = datetime.utcnow()
    TestFSArtifact.write_dummy_partitioned(uid, yesterday)
    TestFSArtifact.write_dummy_partitioned(uid, today)

    r = TestCatalog.get(uid, partition=yesterday, model=FSArtifact)
    if r.files_pattern is None:
        pytest.fail("files_pattern should be set")
    else:
        assert TestFSArtifact.config().fs_prod_prefix() in r.files_pattern
        assert "ala" in Path(r.files_pattern).read_text()
        assert Path(r.files_pattern).parent.joinpath("_SUCCESS").exists()
        manifest_path = Path(r.files_pattern).parent.joinpath("_MANIFEST.json")
        assert manifest_path.exists()
        manifest = FSArtifact.parse_file(manifest_path)
        # Datastore adds tzinfo, which we don't have
        manifest.partition = cast(datetime, manifest.partition).replace(tzinfo=UTC)
        manifest.created = cast(datetime, manifest.created).replace(tzinfo=UTC)
        assert manifest == r

    assert len(list(TestCatalog.lookup(uid))) == 2
    assert len(list(TestCatalog.lookup(uid, partition_dt_start=today))) == 1
    assert len(list(TestCatalog.lookup(uid, partition_dt_start=yesterday))) == 2
    # end is not inclusive
    assert len(list(TestCatalog.lookup(uid, partition_dt_end=today))) == 1


def test_catalog_date_partition(uid: ID) -> None:
    dt = date.today()
    TestFSArtifact.write_dummy_partitioned(uid, dt)
    assert TestCatalog.get(uid, partition=dt).id == uid


def test_catalog_lookup__ordered_partitions(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 10)
    # should return exactly 3 results ordered desc
    _1, _2, _3 = tuple(TestCatalog.lookup(uid, partition_dt_start=date.min, limit=3))
    today = datetime.combine(date.today(), time.min).replace(tzinfo=UTC)
    assert _1.partition == today
    assert _2.partition == today - timedelta(days=1)
    assert _3.partition == today - timedelta(days=2)

    all_parts = tuple(TestCatalog.lookup(uid, partition_dt_start=date.min))
    p = [x.partition for x in all_parts]
    assert len(p) == 10
    assert sorted(cast(List[date], p), reverse=True) == p


def test_catalog_lookup__limit_works(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 3)
    # should return exactly 3 results ordered desc
    (_1,) = tuple(TestCatalog.lookup(uid, limit=1, partition_dt_start=date.min))
    today = datetime.combine(date.today(), time.min).replace(tzinfo=UTC)
    assert _1.partition == today


def test_catalog_lookup__partition_limits(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 5)
    today = datetime.combine(date.today(), time.min).replace(tzinfo=UTC)
    today_m_2 = today - timedelta(days=2)
    today_m_4 = today - timedelta(days=4)
    _1, _2 = tuple(
        TestCatalog.lookup(
            uid, partition_dt_start=today_m_4, partition_dt_end=today_m_2
        )
    )
    assert _1.partition == today_m_4 + timedelta(days=1)
    assert _2.partition == today_m_4

    assert (
        len(
            list(
                TestCatalog.lookup(
                    uid, partition_dt_start=today_m_2, partition_dt_end=today_m_4
                )
            )
        )
        == 0
    )
    assert len(list(TestCatalog.lookup(uid, partition_dt_start=date.min))) == 5
    assert (
        len(
            list(
                TestCatalog.lookup(
                    uid, partition_dt_start=date.min, partition_dt_end=today_m_2
                )
            )
        )
        == 2
    )
    assert len(list(TestCatalog.lookup(uid, partition_dt_end=today_m_2))) == 2
    assert len(list(TestCatalog.lookup(uid, partition_dt_end=today_m_4))) == 0


def test_catalog_lookup__all(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 2)
    assert len(list(TestCatalog.lookup())) >= 2


def test_catalog_lookup__metadata(uid: ID, monkeypatch: MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setenv(EXECUTION_URL_ENV_NAME, "http://wwww.url-to-articat-exe.com")
        write_a_couple_of_partitions(f"{uid}foobar", 2)
        with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.2") as a:
            a.metadata = Metadata(
                arbitrary={"a": "foo"}, schema_fields=["foo", "bar", "baz"]
            )

        with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.1") as a:
            a.metadata = Metadata(
                arbitrary={"a": "foo", "b": "bar"},
                schema_fields=["foo", "bar"],
            )

        with TestFSArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
            a.metadata.arbitrary.update({"a": "foo"})
            a.metadata.schema_fields = ["foo", "bar"]

    assert len(list(TestCatalog.lookup(uid))) == 3
    (_1,) = tuple(TestCatalog.lookup(uid, version="0.1.0"))
    assert _1.metadata and _1.version == "0.1.0"
    assert "git_repo_url" in _1.metadata.arbitrary
    assert "git_head_hash" in _1.metadata.arbitrary
    assert "execution_url" in _1.metadata.arbitrary
    assert (
        _1.metadata.arbitrary["execution_url"] == "http://wwww.url-to-articat-exe.com"
    )
    assert (
        len(tuple(TestCatalog.lookup(uid, metadata=Metadata(schema_fields=["bar"]))))
        == 3
    )
    (_1,) = tuple(TestCatalog.lookup(uid, metadata=Metadata(schema_fields=["baz"])))
    assert _1.metadata and _1.metadata.schema_fields == ["foo", "bar", "baz"]
    assert _1.metadata and _1.version == "0.1.2"
    # there is no version 0.1.0 with baz field
    assert (
        len(
            tuple(
                TestCatalog.lookup(
                    uid, version="0.1.0", metadata=Metadata(schema_fields=["baz"])
                )
            )
        )
        == 0
    )
    assert len(tuple(TestCatalog.lookup(uid, version="0.1.3"))) == 0
    with pytest.raises(ValueError, match="Arbitrary lookup not supported"):
        tuple(TestCatalog.lookup(uid, metadata=Metadata(arbitrary={"a": "foo"})))


def test_catalog_lookup__unordered_when_no_partition_requested(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 10)
    # no partition requested:
    all_parts = tuple(TestCatalog.lookup(uid))
    p = [x.partition for x in all_parts]
    assert len(p) == 10
    assert sorted(cast(List[date], p), reverse=True) != p


def test_catalog_lookup__cant_lookup_by_both_partition_and_metadata(uid: ID) -> None:
    with pytest.raises(
        ValueError, match="Can't lookup based on both metadata and partition"
    ):
        list(
            TestCatalog.lookup(
                id=uid, partition_dt_start=date.today(), metadata=Metadata()
            )
        )


def test_catalog_get__latest_no_part_ver(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 3)
    r = TestCatalog.get(uid)
    assert isinstance(r.partition, datetime)
    assert r.partition.date() == date.today()


def test_catalog_lookup__cant_lookup_by_description_or_spark_schema(uid: ID) -> None:
    with pytest.raises(ValueError, match="Description lookup not supported"):
        list(TestCatalog.lookup(id=uid, metadata=Metadata(description="blah")))
    with pytest.raises(ValueError, match="Spark schema lookup not supported"):
        list(TestCatalog.lookup(id=uid, metadata=Metadata(spark_schema="blah")))


def test_dev_mode_can_overwrite(uid: ID) -> None:
    TestFSArtifact.write_dummy_versioned(uid, "0.1.0")
    with pytest.raises(
        ValueError, match="Catalog already has an entry for this artifact!"
    ):
        TestFSArtifact.write_dummy_versioned(uid, "0.1.0")

    dev_id = f"_dev_{uid}"
    retired = list(
        TestCatalog._lookup(
            dev_id, version="0.1.0", client=TestCatalog._retired_client()
        )
    )
    dev = list(TestCatalog.lookup(dev_id, version="0.1.0"))
    assert len(dev) == 0
    assert len(retired) == 0

    TestFSArtifact.write_dummy_versioned(dev_id, "0.1.0")
    TestFSArtifact.write_dummy_versioned(dev_id, "0.1.0")

    retired = list(
        TestCatalog._lookup(
            dev_id, version="0.1.0", client=TestCatalog._retired_client()
        )
    )
    dev = list(TestCatalog.lookup(dev_id, version="0.1.0"))
    assert len(dev) == 1
    assert len(retired) == 1


def test_catalog_latest_partition(uid: ID) -> None:
    write_a_couple_of_partitions(uid, 10)
    latest = TestCatalog.latest_partition(uid)
    assert (
        max(x.partition for x in TestCatalog.lookup(uid) if x.partition is not None)
        == latest.partition
    )


def test_catalog_get_nice_error_on_missing_get(uid: ID) -> None:
    with pytest.raises(ValueError, match="Can't find requested artifact"):
        TestCatalog.get(uid, version="0.1.1")


def test_catalog_get_nice_error_on_missing_latest(uid: ID) -> None:
    with pytest.raises(ValueError, match="Can't find requested artifact"):
        TestCatalog.latest_partition(uid)


def test_catalog_to_df(uid: ID) -> None:
    TestFSArtifact.write_dummy_partitioned(uid, date.today())
    df = TestCatalog.to_dataframe()
    assert df.id.str.contains(uid).any()
    assert any([c.startswith("metadata_arbitrary_") for c in df.columns])

    df_no_arb = TestCatalog.to_dataframe(include_arbitrary=False)
    assert not any([c.startswith("metadata_arbitrary_") for c in df_no_arb.columns])



import subprocess
import tempfile
from datetime import datetime
from pathlib import Path
from typing import List

import pytest
from _pytest.monkeypatch import MonkeyPatch
from pytest import fixture

from articat.artifact import Artifact
from articat.fs_artifact import FSArtifact
from articat.tests.utils import TestFSArtifact
from articat.utils import utils
from articat.utils.path_utils import cwd
from articat.utils.typing import PathType
from articat.utils.utils import download_artifact, dummy_unsafe_cache, get_repo_and_hash


def get_source_path_that_looks_like_path_from_catalog() -> Path:
    p = Path(
        tempfile.mkdtemp(dir=TestFSArtifact.config().fs_prod_prefix(), suffix="__ID__")
    ).joinpath("__PART__")
    p.mkdir(parents=True)
    return p


def test_happy_path():
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("out-1.txt").write_text("sth-1")
    src.joinpath("out-2.txt").write_text("sth-2")
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/*")
    dst = Path(tempfile.mktemp())
    r = download_artifact(a, dst)
    assert dst.joinpath("out-1.txt").read_text() == "sth-1"
    assert dst.joinpath("out-2.txt").read_text() == "sth-2"
    assert r == f"{dst}/*"


def test_download__empty():
    src = get_source_path_that_looks_like_path_from_catalog()
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/**")
    dst = Path(tempfile.mktemp())
    with pytest.raises(ValueError, match="Nothing to copy in"):
        download_artifact(a, dst)


def test_download__weird_valid_pattern():
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("part-1").write_text("sth-2")
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/**part-1")
    dst = Path(tempfile.mktemp())
    r = download_artifact(a, dst)
    assert dst.joinpath("part-1").read_text() == "sth-2"
    assert r == f"{dst}/**part-1"


def test_download__weird_invalid_pattern():
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("part-1").write_text("sth-2")
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/*/*part-1")
    dst = Path(tempfile.mktemp())
    with pytest.raises(ValueError, match="Nothing to copy"):
        download_artifact(a, dst)


def test_download__single_file():
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("out-1.txt").write_text("sth-1")
    src.joinpath("out-2.txt").write_text("sth-2")
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/out-1.txt")
    dst = Path(tempfile.mktemp())
    r = download_artifact(a, dst)
    assert dst.joinpath("out-1.txt").read_text() == "sth-1"
    assert not dst.joinpath("out-2.txt").exists()
    assert r == f"{dst}/out-1.txt"


def test_download__embedded_dir():
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("out-1.txt").write_text("sth-1")
    src.joinpath("out-2.txt").write_text("sth-2")
    src.joinpath("emb_dir").mkdir()
    src.joinpath("emb_dir").joinpath("out-3.txt").write_text("sth-3")
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/**")
    dst = Path(tempfile.mktemp())
    r = download_artifact(a, dst.as_posix())
    assert dst.joinpath("out-1.txt").read_text() == "sth-1"
    assert dst.joinpath("out-2.txt").read_text() == "sth-2"
    assert dst.joinpath("emb_dir").joinpath("out-3.txt").read_text() == "sth-3"
    assert r == f"{dst}/**"


def test_download__existing_local_dir():
    src = get_source_path_that_looks_like_path_from_catalog()
    a = TestFSArtifact(id="sth", files_pattern=f"{src}/**")
    dst = Path(tempfile.mkdtemp())
    with pytest.raises(ValueError, match="Local directory already exists"):
        download_artifact(a, dst)


@fixture(scope="session")
def caching_artifact() -> FSArtifact:
    src = get_source_path_that_looks_like_path_from_catalog()
    src.joinpath("out-1.txt").write_text("sth-1")
    src.joinpath("out-2.txt").write_text("sth-2")
    return TestFSArtifact(
        id="sth",
        partition=datetime.utcnow(),
        created=datetime.utcnow(),
        files_pattern=f"{src}/out-1.txt",
    )


def test_cache__happy_path(caching_artifact: FSArtifact) -> None:
    cache_dir = Path(tempfile.mkdtemp())
    b = dummy_unsafe_cache(caching_artifact, cache_dir)
    assert b.endswith("out-1.txt")
    assert caching_artifact.files_pattern is not None
    assert caching_artifact.files_pattern not in b
    assert cache_dir.as_posix() in b
    c = dummy_unsafe_cache(caching_artifact, cache_dir)
    assert b == c


def test_cache__env_var(monkeypatch: MonkeyPatch, caching_artifact: FSArtifact) -> None:
    cache_dir = tempfile.mkdtemp()
    diff_cache_dir = tempfile.mkdtemp()
    monkeypatch.setenv("ARTICAT_CACHE_DIR", cache_dir)
    b = dummy_unsafe_cache(caching_artifact, diff_cache_dir)
    assert diff_cache_dir in b
    assert cache_dir not in b
    c = dummy_unsafe_cache(caching_artifact)
    assert b != c
    assert diff_cache_dir not in c
    assert cache_dir in c


def test_cache__diff_artifact_id(caching_artifact: FSArtifact) -> None:
    cache_dir = Path(tempfile.mkdtemp())
    b = dummy_unsafe_cache(caching_artifact, cache_dir)
    diff_art = caching_artifact.copy(update={"id": "sth_diff"})
    c = dummy_unsafe_cache(diff_art, cache_dir)
    assert cache_dir.as_posix() in b
    assert cache_dir.as_posix() in c
    assert b != c


def test_cache__diff_artifact_partition_version_created(
    monkeypatch: MonkeyPatch, caching_artifact: FSArtifact
) -> None:
    monkeypatch.setattr(Artifact, "_partition_str_format", "%Y%m%dT%H%M%S%f")
    cache_dir = Path(tempfile.mkdtemp())
    b = dummy_unsafe_cache(caching_artifact, cache_dir)
    diff_partition = caching_artifact.copy(update={"partition": datetime.utcnow()})
    diff_version = caching_artifact.copy(update={"version": "0.1.3"})
    diff_created = caching_artifact.copy(update={"created": datetime.utcnow()})
    c = dummy_unsafe_cache(diff_partition, cache_dir)
    d = dummy_unsafe_cache(diff_version, cache_dir)
    e = dummy_unsafe_cache(diff_created, cache_dir)

    assert all(cache_dir.as_posix() in x for x in [b, c, d, e])
    assert b != c != d != e


def test_cache__download_failure_scenario(
    caching_artifact: FSArtifact, monkeypatch: MonkeyPatch
) -> None:
    cache_dir = tempfile.mkdtemp()

    def raising_download(artifact: FSArtifact, local_dir: PathType) -> str:
        # we download everything
        str = download_artifact(artifact, local_dir)
        # and then simulate an error
        assert Path(str).exists()
        # we make sure at this stage cache dir is not empty
        assert cache_dir in str
        assert len(list(Path(cache_dir).glob("*"))) > 0
        raise ValueError("blah")

    monkeypatch.setattr(utils, "download_artifact", raising_download)
    with pytest.raises(ValueError, match="Cache download failed, make sure"):
        dummy_unsafe_cache(caching_artifact, cache_dir)

    assert len(list(Path(cache_dir).glob("*"))) == 0


def test_git__get_repo_curr_repo():
    repo_url, hash = get_repo_and_hash()
    assert "related-sciences/articat" in repo_url
    assert len(hash) > 0


def test_git__dirty_tree():
    def run_git(cmd: List[str], out: int = subprocess.DEVNULL) -> None:
        subprocess.check_call(["git"] + cmd, stderr=out, stdout=out)

    repo_dir = tempfile.mkdtemp()
    with cwd(repo_dir):
        run_git(["init"])
        run_git(["config", "user.email", "foo@example.com"])
        run_git(["config", "user.name", "Foo Bar"])
        run_git(
            [
                "commit",
                "--allow-empty",
                "-m foo_commit",
            ]
        )
        run_git(["remote", "add", "origin", "git@github.com:foo/bar.git"])

        _, hash = get_repo_and_hash()
        assert "DIRTY" not in hash

        some_file = Path(repo_dir).joinpath("some_file")
        some_file.touch()
        _, hash = get_repo_and_hash()
        assert hash.endswith("DIRTY")
        some_file.unlink()
        _, hash = get_repo_and_hash()
        assert "DIRTY" not in hash

        some_file.touch()
        run_git(["add", some_file.as_posix()])
        _, hash = get_repo_and_hash()
        assert hash.endswith("DIRTY")

        run_git(["commit", "-m bar_commit"])
        repo_url, hash = get_repo_and_hash()
        assert "DIRTY" not in hash
        assert repo_url == "git@github.com:foo/bar.git"



import inspect
import logging
import os
import re
import shutil
import subprocess
from hashlib import md5
from pathlib import Path
from typing import Optional, Tuple

import fsspec
from fsspec import AbstractFileSystem

from articat.artifact import Artifact
from articat.fs_artifact import FSArtifact
from articat.utils.typing import PathType

logger = logging.getLogger(__name__)


def _git_tree_is_dirty() -> bool:
    status_out = subprocess.check_output(
        ["git", "status", "--short", "--untracked-files=all", "--ignored=no"]
    )
    # NOTE: if status out is empty, tree is clean
    return len(status_out) > 0


def _git_get_head_hash() -> str:
    return (
        subprocess.check_output(["git", "rev-parse", "--verify", "HEAD"])
        .decode()
        .strip()
    )


def _git_get_remote_url(remote: str) -> str:
    return (
        subprocess.check_output(["git", "remote", "get-url", remote]).decode().strip()
    )


def get_repo_and_hash(remote_location: Optional[str] = None) -> Tuple[str, str]:
    """
    Returns git remote repository URL and HEAD hash. If the tree at
    HEAD is dirty, the hash will include "-DIRTY" suffix. Expect the CWD
    to be within the git repository.
    """
    if _git_tree_is_dirty():
        head_hash = f"{_git_get_head_hash()}-DIRTY"
    else:
        head_hash = _git_get_head_hash()

    remote_location = (
        remote_location or os.environ.get("ARTICAT_GIT_REMOTE") or "origin"
    )
    return _git_get_remote_url(remote_location), head_hash


def get_call_site(frames_back: int) -> Optional[Tuple[str, int]]:
    """
    Returns path of the call site file, and line number

    :param frames_back: number of frames to look back in the call stack for the
                        call site of interest. For example 1 is the call site of
                        the get_relative_call_site itself, 2 is the call site of
                        the function that called get_relative_call_site and so on.
    """
    caller_frame = inspect.stack()[frames_back]
    caller_fname_path = Path(caller_frame.filename)
    caller_lineno = caller_frame.lineno
    return caller_fname_path.as_posix(), caller_lineno


def download_artifact(artifact: FSArtifact, local_dir: PathType) -> str:
    """
    Downloads artifact outputs to local directory.
    """
    if isinstance(local_dir, str):
        local_dir = Path(local_dir)
    if local_dir.exists():
        raise ValueError(
            "Local directory already exists, please point at a non existing directory"
        )
    local_dir.mkdir(parents=True)
    protocol = artifact._get_protocol(artifact.files_pattern)
    src_fs: AbstractFileSystem = fsspec.filesystem(protocol)
    assert artifact.files_pattern is not None
    assert artifact.main_dir is not None
    prefix = artifact.main_dir
    # Note: glob results don't have fs scheme
    prefix_no_scheme = re.sub(FSArtifact._fs_scheme_regex, "", prefix)
    to_copy = src_fs.glob(artifact.files_pattern)
    if len(to_copy) == 0:
        raise ValueError(f"Nothing to copy in `{artifact.files_pattern}`")
    for f in to_copy:
        assert (
            prefix_no_scheme in f
        ), f"prefix {prefix_no_scheme}, should be present in the file path {f}"
        dst = f.replace(prefix_no_scheme, local_dir.as_posix())
        if src_fs.isdir(f):
            Path(dst).mkdir()
        else:
            src_fs.download(f, dst)
    suffix = artifact.files_pattern.replace(prefix, "")
    return f"{local_dir.as_posix()}{suffix}"


def dummy_unsafe_cache(
    artifact: FSArtifact, cache_dir: Optional[PathType] = None
) -> str:
    """
    This is an unsafe implementation of local FS caching for FSArtifacts.

    Context:
    We don't want to build caching into articat, since it's a hard problem to
    solve properly. This is a helper method, and should be used at your
    own risk (and definitely not in the production environment).

    Example:

    ```
    a = Catalog.latest_partition("id")
    df = spark.read.text(dummy_unsafe_cache(a))
    ```

    :param artifact: FSArtifact to return data for
    :param cache_dir: cache dir location, can be provided either by this argument
        or environment variable ARTICAT_CACHE_DIR. Argument takes precedence over
        environment variable.
    :return: string representing the local version of `artifact.files_pattern`
    """
    if cache_dir is None:
        cache_dir = os.environ.get("ARTICAT_CACHE_DIR")
        if cache_dir is not None:
            cache_dir = Path(cache_dir)
        else:
            raise ValueError(
                "No cache_dir set and no ARTICAT_CACHE_DIR env variable, don't know where to store cache"
            )
    assert cache_dir is not None
    cache_dir_path: Path = Path(cache_dir) if isinstance(cache_dir, str) else cache_dir
    if not cache_dir_path.exists():
        logger.info(f"Creating cache dir: {cache_dir_path.as_posix()}")
        cache_dir_path.mkdir(parents=True)
    m = md5()
    assert artifact.id is not None
    m.update(artifact.id.encode("UTF-8"))
    assert artifact.partition is not None
    m.update(
        artifact.partition.strftime(Artifact._partition_str_format).encode("UTF-8")
    )
    assert artifact.created is not None
    m.update(artifact.created.strftime(Artifact._partition_str_format).encode("UTF-8"))
    m.update((artifact.version and artifact.version.encode("UTF-8")) or b"")
    hex = m.hexdigest()

    cache_entry = cache_dir_path.joinpath(hex)
    if cache_entry.exists():
        logger.info(f"Cache hit for artifact {artifact.id}, returning local data ...")
        # TODO (rav): use main_dir when it's available
        assert artifact.files_pattern is not None
        assert artifact.main_dir is not None
        suffix = artifact.files_pattern.replace(artifact.main_dir, "")
        return f"{cache_entry.as_posix()}{suffix}"
    else:
        logger.info(f"Cache miss for artifact {artifact.id}, downloading data ...")
        try:
            return download_artifact(artifact, cache_entry)
        except BaseException as e:
            logger.exception(
                f"Cache download failed, will do my best to invalidate cache entry {cache_entry.as_posix()}"
            )
            try:
                shutil.rmtree(cache_entry)
            finally:
                raise ValueError(
                    f"Cache download failed, make sure {cache_entry.as_posix()} does not exits!"
                ) from e



from datetime import date, datetime, time
from pathlib import Path
from typing import Any, Iterator

from _pytest.capture import CaptureFixture
from _pytest.monkeypatch import MonkeyPatch

from articat.artifact import Artifact, Metadata
from articat.catalog import Catalog
from articat.cli import CLI
from articat.fs_artifact import FSArtifact
from articat.tests.utils import TestFSArtifact


def get_dummy_artifact() -> Artifact:
    dt = datetime.combine(date(1642, 12, 25), time.min)
    return FSArtifact(
        id="sth",
        partition=dt,
        created=dt,
        version="0.1.0",
        metadata=Metadata(),
        files_pattern="/tmp/bah",
    )


def get_lookup(a: Artifact):
    def my_lookup(**_: Any) -> Iterator[Artifact]:
        yield a

    return my_lookup


def test_happy_path(monkeypatch: MonkeyPatch, capsys: CaptureFixture) -> None:
    monkeypatch.setattr(Catalog, "lookup", get_lookup(get_dummy_artifact()))
    CLI.lookup()
    out, _ = capsys.readouterr()
    expected = '"files_pattern": "/tmp/bah"'
    assert expected in out
    assert "metadata" not in out
    CLI.lookup(no_trunc=True)
    out, _ = capsys.readouterr()
    assert "metadata" in out


def test_open_artifact_id(monkeypatch: MonkeyPatch, caplog: CaptureFixture) -> None:
    a = get_dummy_artifact()

    def dummy_open_browser():
        print(f"Would open browser for {a.spec()}")

    object.__setattr__(a, "open_browser", dummy_open_browser)
    monkeypatch.setattr(Catalog, "lookup", get_lookup(a))
    CLI.open(id="sth")
    assert f"Opening artifact: {a.spec()}" in caplog.text


def test_local_dump(tmpdir: str) -> None:
    tmpdir_path = Path(tmpdir).joinpath("foo")
    tmpdir_path.write_text("txt")
    desc = "description"
    a = CLI.dump(
        id="test",
        local_path=tmpdir_path.as_posix(),
        version="0.1.0",
        description=desc,
        dev=True,
        _test_artifact=TestFSArtifact.partitioned(id="test", dev=True),
    )
    assert a.metadata.description == desc
    assert Path(a.joinpath("foo")).read_text() == "txt"



from pathlib import Path
from typing import ClassVar

import pytest
from pytest import fixture

from articat.artifact import ID
from articat.notebook_artifact import NotebookArtifact
from articat.tests.utils import TestCatalog, TestFSArtifactMixin

pytestmark = pytest.mark.datastore_emulated

FIXTURE_DIR = Path(__file__).resolve().parent.joinpath("test_files")


@fixture
def dummy_notebook(tmpdir: str):
    return FIXTURE_DIR.joinpath("dummy_notebook.ipynb")


class TestNotebookArtifact(TestFSArtifactMixin, NotebookArtifact):
    """NotebookArtifact class for testing purposes"""

    __test__: ClassVar[bool] = False


def test_notebook_artifact(uid: ID, dummy_notebook: Path) -> None:
    with TestNotebookArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.stage_notebook(dummy_notebook, params={"a": 3})
        a.files_pattern = None

    out = a.fetch(TestCatalog)
    assert Path(out.joinpath("output.html")).stat().st_size > 0


def test_notebook_artifact_custom(uid: ID, tmpdir) -> None:
    custom_notebook_html = Path(tmpdir).joinpath("report_foo.html")
    html = "<html>FOO</html>"
    custom_notebook_html.write_text(html)
    with TestNotebookArtifact.dummy_versioned_ctx(uid, "0.1.0") as a:
        a.stage_rendered_notebook(custom_notebook_html.as_posix(), filename=None)
        a.files_pattern = None

    out = a.fetch(TestCatalog)
    assert Path(out.joinpath("report_foo.html")).read_text() == html
